from django.contrib import admin

from .models import Message, Dispatch, DispatchError, Subscription


class DispatchInlineAdmin(admin.TabularInline):

    model = Dispatch
    extra = 0
    raw_id_fields = ('recipient',)
    readonly_fields = ('retry_count',)


class DispatchErrorInlineAdmin(admin.TabularInline):

    model = DispatchError
    extra = 0
    readonly_fields = ('time_created', 'dispatch', 'error_log')


class MessageAdmin(admin.ModelAdmin):

    list_display = ('time_created', 'cls', 'dispatches_ready')
    list_filter = ('cls', 'dispatches_ready')
    ordering = ('-time_created',)

    inlines = (DispatchInlineAdmin,)


class DispatchAdmin(admin.ModelAdmin):

    list_display = ('time_created', 'dispatch_status', 'address', 'time_dispatched', 'messenger', 'retry_count')
    list_filter = ('dispatch_status', 'messenger')
    ordering = ('-time_created',)
    raw_id_fields = ('recipient',)
    readonly_fields = ('retry_count',)

    inlines = (DispatchErrorInlineAdmin,)


class DispatchErrorAdmin(admin.ModelAdmin):

    list_display = ('time_created', 'dispatch')
    ordering = ('-time_created',)
    readonly_fields = ('time_created', 'dispatch', 'error_log')


class SubscriptionAdmin(admin.ModelAdmin):

    list_display = ('time_created', 'message_cls', 'messenger_cls')
    ordering = ('-time_created',)
    readonly_fields = ('time_created',)
    list_filter = ('message_cls', 'messenger_cls')


admin.site.register(Message, MessageAdmin)
admin.site.register(Dispatch, DispatchAdmin)
admin.site.register(DispatchError, DispatchErrorAdmin)
admin.site.register(Subscription, SubscriptionAdmin)



from django.shortcuts import redirect
from django.contrib.staticfiles.templatetags.staticfiles import static as get_static_url

from .models import Dispatch
from .exceptions import UnknownMessageTypeError
from .signals import sig_unsubscribe_failed, sig_mark_read_failed


def _generic_view(message_method, fail_signal, request, message_id, dispatch_id, hashed, redirect_to=None):

    if redirect_to is None:
        redirect_to = '/'

    try:
        dispatch = Dispatch.objects.select_related('message').get(pk=dispatch_id)
        if int(message_id) != dispatch.message_id:
            raise ValueError()
        message = dispatch.message
    except (Dispatch.DoesNotExist, ValueError):
        pass
    else:
        try:
            message_type = message.get_type()
            expected_hash = message_type.get_dispatch_hash(dispatch_id, message_id)

            method = getattr(message_type, message_method)
            return method(
                request, message, dispatch,
                hash_is_valid=(expected_hash == hashed),
                redirect_to=redirect_to
            )
        except UnknownMessageTypeError:
            pass

    fail_signal.send(None, request=request, message=message_id, dispatch=dispatch_id)

    return redirect(redirect_to)


def unsubscribe(request, message_id, dispatch_id, hashed, redirect_to=None):
    """Handles unsubscribe request.

    :param Request request:
    :param int message_id:
    :param int dispatch_id:
    :param str hashed:
    :param str redirect_to:
    :return:
    """
    return _generic_view(
        'handle_unsubscribe_request', sig_unsubscribe_failed,
        request, message_id, dispatch_id, hashed, redirect_to=redirect_to
    )


def mark_read(request, message_id, dispatch_id, hashed, redirect_to=None):
    """Handles mark message as read request.

    :param Request request:
    :param int message_id:
    :param int dispatch_id:
    :param str hashed:
    :param str redirect_to:
    :return:
    """
    if redirect_to is None:
        redirect_to = get_static_url('img/sitemessage/blank.png')

    return _generic_view(
        'handle_mark_read_request', sig_mark_read_failed,
        request, message_id, dispatch_id, hashed, redirect_to=redirect_to
    )



from contextlib import contextmanager

from ..utils import Recipient, is_iterable, get_registered_message_type
from ..models import Dispatch
from ..exceptions import UnknownMessageTypeError


class MessengerBase(object):
    """Base class for messengers used by sitemessage.

    Custom messenger classes, implementing various message delivery
    mechanics, other messenger classes must inherit from this one.

    """

    # Messenger alias to address it from different places, Should rather be quite unique %)
    alias = None
    # Title to show to user.
    title = None

    # Makes subscription for this messenger messages available for users (see get_user_preferences_for_ui())
    allow_user_subscription = True

    # Dispatches by status dict will be here runtime. See init_delivery_statuses_dict().
    _st = None

    @classmethod
    def get_alias(cls):
        """Returns messenger alias.

        :return: str
        :rtype: str
        """
        if cls.alias is None:
            cls.alias = cls.__name__
        return cls.alias

    def __str__(self):
        return self.__class__.get_alias()

    @contextmanager
    def before_after_send_handling(self):
        """Context manager that allows to execute send wrapped
        in before_send() and after_send().

        """
        self.before_send()
        yield
        self.after_send()

    def send_test_message(self, to, text):
        """Sends a test message using messengers settings.

        :param str to: an address to send test message to
        :param str text: text to send
        """
        with self.before_after_send_handling():
            result = self._test_message(to, text)
        return result

    def _test_message(self, to, text):
        """This method should be implemented by a heir to send a test message.

        :param str to: an address to send test message to
        :param str text: text to send
        """
        raise NotImplementedError(self.__class__.__name__ + ' must implement `test_message()`.')

    @classmethod
    def get_address(cls, recipient):
        """Returns recipient address.

        Heirs may override this to deduce address from `recipient` data
        (e.g. to get address from Django User model instance).

        :param object recipient: any object passed to `recipients()`
        :return: str
        :rtype: str
        """
        return recipient

    @classmethod
    def _structure_recipients_data(cls, recipients):
        """Converts recipients data into a list of Recipient objects.

        :param list recipients: list of objects
        :return: list of Recipient
        :rtype: list
        """
        try:  # That's all due Django 1.7 apps loading.
            from django.contrib.auth import get_user_model
            USER_MODEL = get_user_model()
        except ImportError:
            # Django 1.4 fallback.
            from django.contrib.auth.models import User as USER_MODEL

        if not is_iterable(recipients):
            recipients = (recipients,)

        objects = []
        for r in recipients:
            user = None
            if isinstance(r, USER_MODEL):
                user = r
            address = cls.get_address(r)  # todo maybe raise an exception of not a string?

            objects.append(Recipient(cls.get_alias(), user, address))

        return objects

    def _init_delivery_statuses_dict(self):
        """Initializes a dict indexed by message delivery statuses."""
        self._st = {
            'pending': [],
            'sent': [],
            'error': [],
            'failed': []
        }

    def mark_pending(self, dispatch):
        """Marks a dispatch as pending.

        Should be used within send().

        :param Dispatch dispatch: a Dispatch
        """
        self._st['pending'].append(dispatch)

    def mark_sent(self, dispatch):
        """Marks a dispatch as successfully sent.

        Should be used within send().

        :param Dispatch dispatch: a Dispatch
        """
        self._st['sent'].append(dispatch)

    def mark_error(self, dispatch, error_log, message_cls):
        """Marks a dispatch as having error or consequently as failed
        if send retry limit for that message type is exhausted.

        Should be used within send().

        :param Dispatch dispatch: a Dispatch
        :param str error_log: error message
        :param MessageBase message_cls: MessageBase heir
        """
        if message_cls.send_retry_limit is not None and (dispatch.retry_count + 1) >= message_cls.send_retry_limit:
            self.mark_failed(dispatch, error_log)
        else:
            dispatch.error_log = error_log
            self._st['error'].append(dispatch)

    def mark_failed(self, dispatch, error_log):
        """Marks a dispatch as failed.

        Sitemessage won't try to deliver already failed messages.

        Should be used within send().

        :param Dispatch dispatch: a Dispatch
        :param str error_log: str - error message
        """
        dispatch.error_log = error_log
        self._st['failed'].append(dispatch)

    def before_send(self):
        """This one is called right before send procedure.
        Usually heir will implement some messenger warm up (connect) code.

        """

    def after_send(self):
        """This one is called right after send procedure.
        Usually heir will implement some messenger cool down (disconnect) code.

        """

    def _process_messages(self, messages, ignore_unknown_message_types=False):
        """Performs message processing.

        :param dict messages: indexed by message id dict with messages data
        :param bool ignore_unknown_message_types: whether to silence exceptions
        :raises UnknownMessageTypeError:
        """
        self._init_delivery_statuses_dict()

        with self.before_after_send_handling():
            for message_id, message_data in messages.items():
                message_model, dispatch_models = message_data
                try:
                    message_cls = get_registered_message_type(message_model.cls)
                except UnknownMessageTypeError:
                    if ignore_unknown_message_types:
                        continue
                    raise

                message_type_cache = None
                for dispatch in dispatch_models:
                    if not dispatch.message_cache:  # Create actual message text for further usage.

                        if message_type_cache is None and not message_cls.has_dynamic_context:
                            # If a message class doesn't depend upon a dispatch data for message compilation,
                            # we'd compile a message just once.
                            message_type_cache = message_cls.compile(message_model, self, dispatch=dispatch)

                        dispatch.message_cache = message_type_cache or message_cls.compile(
                            message_model, self, dispatch=dispatch
                        )

                self.send(message_cls, message_model, dispatch_models)

        self._update_dispatches()

    def _update_dispatches(self):
        """Updates dispatched data in DB according to information gather by `mark_*` methods,"""
        Dispatch.log_dispatches_errors(self._st['error'] + self._st['failed'])
        Dispatch.set_dispatches_statuses(**self._st)
        self._init_delivery_statuses_dict()

    def send(self, message_cls, message_model, dispatch_models):
        """Main send method must be implement by all heirs.

        :param MessageBase message_cls: a MessageBase heir
        :param Message message_model: message model
        :param list dispatch_models: Dispatch models for this Message
        """
        raise NotImplementedError(self.__class__.__name__ + ' must implement `send()`.')



from django.conf import settings


# Module name to search sitemessage preferences in.
APP_MODULE_NAME = getattr(settings, 'SITEMESSAGE_APP_MODULE_NAME', 'sitemessages')

# Whether to register builtin message types.
INIT_BUILTIN_MESSAGE_TYPES = getattr(settings, 'SITEMESSAGE_INIT_BUILTIN_MESSAGE_TYPES', True)

# Priority for messages sent by Django Email backend (sitemessage.backends.EmailBackend).
EMAIL_BACKEND_MESSAGES_PRIORITY = getattr(settings, 'SITEMESSAGE_EMAIL_BACKEND_MESSAGES_PRIORITY', None)

# Message type alias for messages sent `schedule_email` shortcut.
DEFAULT_SHORTCUT_EMAIL_MESSAGES_TYPE = getattr(settings, 'SITEMESSAGE_DEFAULT_SHORTCUT_EMAIL_MESSAGES_TYPE', 'smtp')

# Site URL to use in messages.
SITE_URL = getattr(settings, 'SITEMESSAGE_SITE_URL', None)



from django.utils.translation import ugettext as _

from .base import MessengerBase
from ..exceptions import MessengerWarmupException, MessengerException


class TelegramMessengerException(MessengerException):
    """Exceptions raised by Telegram messenger."""


class TelegramMessenger(MessengerBase):
    """Implements Telegram message delivery via Telegram Bot API.

    Uses `requests` module: https://pypi.python.org/pypi/requests

    """

    alias = 'telegram'
    title = _('Telegram')

    _session_started = False
    _tpl_url = 'https://api.telegram.org/bot%(token)s/%(method)s'

    def __init__(self, auth_token):
        """Configures messenger.

        Register a Telegram Bot using instructions from https://core.telegram.org/bots/api

        :param auth_token: Bot unique authentication token
        """
        import requests

        self.lib = requests
        self.auth_token = auth_token

    @classmethod
    def get_address(cls, recipient):
        return getattr(recipient, 'telegram', None) or recipient

    def _test_message(self, to, text):
        return self._send_message(self._build_message(to, text))

    def _verify_bot(self):
        """Sends an API command to test whether bot is authorized."""
        self._send_command('getMe')

    def get_updates(self):
        """Returns new messages addressed to bot.

        :rtype: list
        """
        with self.before_after_send_handling():
            result = self._send_command('getUpdates')
        return result

    def get_chat_ids(self):
        """Returns unique chat IDs from `/start` command messages sent to our bot by users.
        Those chat IDs can be used to send messages to chats.

        :rtype: list
        """
        updates = self.get_updates()
        chat_ids = []
        if updates:
            for update in updates:
                message = update['message']
                if message['text'] == '/start':
                    chat_ids.append(message['chat']['id'])
        return list(set(chat_ids))

    def before_send(self):
        if self._session_started:
            return

        try:
            self._verify_bot()
            self._session_started = True
        except TelegramMessengerException as e:
            raise MessengerWarmupException('Telegram Error: %s' % e)

    @classmethod
    def _build_message(cls, to, text):
        return {'chat_id': to, 'text': text}

    def _send_command(self, method_name, data=None):
        """Sends a command to API.

        :param str method_name:
        :param dict data:
        :return:
        """
        try:
            response = self.lib.post(self._tpl_url % {'token': self.auth_token, 'method': method_name}, data=data)
            json = response.json()

            if not json['ok']:
                raise TelegramMessengerException(json['description'])

            return json['result']

        except self.lib.exceptions.RequestException as e:
            raise TelegramMessengerException(e)

    def _send_message(self, msg):
        return self._send_command('sendMessage', msg)

    def send(self, message_cls, message_model, dispatch_models):
        if self._session_started:
            for dispatch_model in dispatch_models:
                msg = self._build_message(dispatch_model.address, dispatch_model.message_cache)
                try:
                    self._send_message(msg)
                    self.mark_sent(dispatch_model)
                except Exception as e:
                    self.mark_error(dispatch_model, e, message_cls)



from .base import MessageBase

from django.utils.translation import ugettext as _


class PlainTextMessage(MessageBase):
    """Simple plain text message class to allow schedule_messages()
    to accept message as a simple string instead of a message object.

    """

    alias = 'plain'
    title = _('Text notification')

    def __init__(self, text):
        super(PlainTextMessage, self).__init__({self.SIMPLE_TEXT_ID: text})



"""This file contains signals emitted by sitemessage."""
import django.dispatch


# Emitted when user unsubscribe requested is successful.
sig_unsubscribe_success = django.dispatch.Signal(providing_args=['request', 'message', 'dispatch'])

# Emitted when user unsubscribe requested fails.
sig_unsubscribe_failed = django.dispatch.Signal(providing_args=['request', 'message', 'dispatch'])

# Emitted when mark read requested is successful.
sig_mark_read_success = django.dispatch.Signal(providing_args=['request', 'message', 'dispatch'])

# Emitted when mark read requested fails.
sig_mark_read_failed = django.dispatch.Signal(providing_args=['request', 'message', 'dispatch'])



from django.core.mail.backends.base import BaseEmailBackend

from .shortcuts import schedule_email
from .settings import EMAIL_BACKEND_MESSAGES_PRIORITY


class EmailBackend(BaseEmailBackend):
    """Email backend for Django built-in mailing functions scheduling messages."""

    def send_messages(self, email_messages):
        if not email_messages:
            return

        sent = 0
        for message in email_messages:

            if not message.recipients():
                continue

            schedule_email(
                {'contents': message.body},
                message.recipients(),
                subject=message.subject, priority=EMAIL_BACKEND_MESSAGES_PRIORITY
            )

            sent += 1

        return sent



from django.utils.translation import ugettext as _

from .base import MessengerBase
from ..exceptions import MessengerWarmupException


class XMPPSleekMessenger(MessengerBase):
    """Implements XMPP message delivery using `sleekxmpp` module.

    http://sleekxmpp.com/

    """

    alias = 'xmppsleek'
    xmpp = None
    title = _('XMPP')
    _session_started = False

    def __init__(self, from_jid, password, host='localhost', port=5222, use_tls=True, use_ssl=False):
        """Configures messenger.

        :param str from_jid: Jabber ID to send messages from
        :param str password: password to log into XMPP server
        :param str host: XMPP server host
        :param int port: XMPP server port
        :param bool use_tls: whether to use TLS
        :param bool use_ssl: whether to use SSL
        :return:
        """
        import sleekxmpp

        self.lib = sleekxmpp

        self.from_jid = from_jid
        self.password = password
        self.host = host
        self.port = port
        self.use_tls = use_tls
        self.use_ssl = use_ssl

    @classmethod
    def get_address(cls, recipient):
        return getattr(recipient, 'jabber', None) or recipient

    def _test_message(self, to, text):
        return self._send_message(to, text)

    def _send_message(self, to, text):
        return self.xmpp.send_message(mfrom=self.from_jid, mto=to, mbody=text, mtype='chat')

    def before_send(self):
        def on_session_start(event):
            try:
                self.xmpp.send_presence()
                self.xmpp.get_roster()
                self._session_started = True
            except self.lib.exceptions.XMPPError as e:
                raise MessengerWarmupException('XMPP Error: %s' % e)

        self.xmpp = self.lib.ClientXMPP(self.from_jid, self.password)
        self.xmpp.add_event_handler('session_start', on_session_start)

        result = self.xmpp.connect(address=(self.host, self.port), reattempt=False, use_tls=self.use_tls, use_ssl=self.use_ssl)
        if result:
            self.xmpp.process(block=False)

    def after_send(self):
        if self._session_started:
            self.xmpp.disconnect(wait=True)  # Wait for a send queue.
            self._session_started = False

    def send(self, message_cls, message_model, dispatch_models):
        if self._session_started:
            for dispatch_model in dispatch_models:
                try:
                    self._send_message(dispatch_model.address, dispatch_model.message_cache)
                    self.mark_sent(dispatch_model)
                except Exception as e:
                    self.mark_error(dispatch_model, e, message_cls)



from collections import OrderedDict, defaultdict
from operator import itemgetter

from django import VERSION
from django.utils import six
from django.conf.urls import patterns, url

from .models import Message, Dispatch, Subscription
from .exceptions import UnknownMessengerError, UnknownMessageTypeError
from .messages.plain import PlainTextMessage
from .views import mark_read, unsubscribe

# NB: Some of these unused imports are exposed as part of toolbox API.
from .messages import register_builtin_message_types
from .utils import (
    is_iterable, import_project_sitemessage_modules, get_site_url, recipients,
    register_messenger_objects, get_registered_messenger_object, get_registered_messenger_objects,
    register_message_types, get_registered_message_type, get_registered_message_types,
    get_message_type_for_app, override_message_type_for_app,
)


_ALIAS_SEP = '|'
_PREF_POST_KEY = 'sm_user_pref'


if VERSION < (1, 7, 0):
    # Trying import sitemessage settings files from project apps.
    import_project_sitemessage_modules()


def schedule_messages(messages, recipients=None, sender=None, priority=None):
    """Schedules a message or messages.

    :param MessageBase|str|list messages: str or MessageBase heir or list - use str to create PlainTextMessage.
    :param list|None recipients: recipients addresses or Django User model heir instances
        If `None` Dispatches should be created before send using `prepare_dispatches()`.
    :param User|None sender: User model heir instance
    :param int priority: number describing message priority. If set overrides priority provided with message type.
    :return: list of tuples - (message_model, dispatches_models)
    :rtype: list
    """
    if not is_iterable(messages):
        messages = (messages,)

    results = []
    for message in messages:
        if isinstance(message, six.string_types):
            message = PlainTextMessage(message)

        resulting_priority = message.priority
        if priority is not None:
            resulting_priority = priority
        results.append(message.schedule(sender=sender, recipients=recipients, priority=resulting_priority))

    return results


def send_scheduled_messages(priority=None, ignore_unknown_messengers=False, ignore_unknown_message_types=False):
    """Sends scheduled messages.

    :param int, None priority: number to limit sending message by this priority.
    :param bool ignore_unknown_messengers: to silence UnknownMessengerError
    :param bool ignore_unknown_message_types: to silence UnknownMessageTypeError
    :raises UnknownMessengerError:
    :raises UnknownMessageTypeError:
    """
    dispatches_by_messengers = Dispatch.group_by_messengers(Dispatch.get_unsent(priority=priority))

    for messenger_id, messages in dispatches_by_messengers.items():
        try:
            messenger_obj = get_registered_messenger_object(messenger_id)
            messenger_obj._process_messages(messages, ignore_unknown_message_types=ignore_unknown_message_types)
        except UnknownMessengerError:
            if ignore_unknown_messengers:
                continue
            raise


def prepare_dispatches():
    """Automatically creates dispatches for messages without them.

    :return: list of Dispatch
    :rtype: list
    """
    dispatches = []
    target_messages = Message.get_without_dispatches()

    cache = {}

    for message_model in target_messages:

        if message_model.cls not in cache:
            message_cls = get_registered_message_type(message_model.cls)
            subscribers = message_cls.get_subscribers()
            cache[message_model.cls] = (message_cls, subscribers)
        else:
            message_cls, subscribers = cache[message_model.cls]

        dispatches.extend(message_cls.prepare_dispatches(message_model))

    return dispatches


def get_user_preferences_for_ui(user, message_filter=None, messenger_filter=None, new_messengers_titles=None):
    """Returns a two element tuple with user subscription preferences to render in UI.

    Message types with the same titles are merged into one row.

    First element:
        A list of messengers titles.

    Second element:
        User preferences dictionary indexed by message type titles.
        Preferences (dictionary values) are lists of tuples:
            (preference_alias, is_supported_by_messenger_flag, user_subscribed_flag)

        Example:
            {'My message type': [('test_message|smtp', True, False), ...]}

    :param User user:
    :param callable|None message_filter: A callable accepting a message object to filter out message types
    :param callable|None messenger_filter: A callable accepting a messenger object to filter out messengers
    :return:
    """
    if new_messengers_titles is None:
        new_messengers_titles = {}

    msgr_to_msg = defaultdict(set)
    msg_titles = OrderedDict()
    msgr_titles = OrderedDict()

    for msgr in get_registered_messenger_objects().values():
        if not (messenger_filter is None or messenger_filter(msgr)) or not msgr.allow_user_subscription:
            continue

        msgr_alias = msgr.alias
        msgr_title = new_messengers_titles.get(msgr.alias) or msgr.title

        for msg in get_registered_message_types().values():
            if not (message_filter is None or message_filter(msg)) or not msg.allow_user_subscription:
                continue

            msgr_supported = msg.supported_messengers
            is_supported = (not msgr_supported or msgr.alias in msgr_supported)

            if not is_supported:
                continue

            msg_alias = msg.alias
            msg_titles.setdefault('%s' % msg.title, []).append(msg_alias)

            msgr_to_msg[msgr_alias].update((msg_alias,))
            msgr_titles[msgr_title] = msgr_alias

    def sort_titles(titles):
        return OrderedDict(sorted([(k, v) for k, v in titles.items()], key=itemgetter(0)))

    msgr_titles = sort_titles(msgr_titles)

    user_prefs = OrderedDict()

    user_subscriptions = ['%s%s%s' % (pref.message_cls, _ALIAS_SEP, pref.messenger_cls)
                          for pref in Subscription.get_for_user(user)]

    for msg_title, msg_aliases in sort_titles(msg_titles).items():
        for __, msgr_alias in msgr_titles.items():
            msg_candidates = msgr_to_msg[msgr_alias].intersection(msg_aliases)

            alias = ''
            msg_supported = False
            subscribed = False

            if msg_candidates:
                alias = '%s%s%s' % (msg_candidates.pop(), _ALIAS_SEP, msgr_alias)
                msg_supported = True
                subscribed = alias in user_subscriptions

            user_prefs.setdefault(msg_title, []).append((alias, msg_supported, subscribed))

    return msgr_titles.keys(), user_prefs


def set_user_preferences_from_request(request):
    """Sets user subscription preferences using data from a request.

    Expects data sent by form built with `sitemessage_prefs_table` template tag.

    :param request:
    :rtype: bool
    :return: Flag, whether prefs were found in the request.
    """
    prefs = []
    for pref in request.POST.getlist(_PREF_POST_KEY):
        message_alias, messenger_alias = pref.split(_ALIAS_SEP)
        try:
            get_registered_message_type(message_alias)
            get_registered_messenger_object(messenger_alias)
        except (UnknownMessengerError, UnknownMessageTypeError):
            pass
        else:
            prefs.append((message_alias, messenger_alias))

    Subscription.replace_for_user(request.user, prefs)

    return bool(prefs)


def get_sitemessage_urls():
    """Returns sitemessage urlpatterns, that can be attached to urlpatterns of a project:

        # Example from urls.py.

        from sitemessage.toolbox import get_sitemessage_urls

        urlpatterns = patterns('',
            # Your URL Patterns belongs here.

        ) + get_sitemessage_urls()  # Now attaching additional URLs.

    """
    url_unsubscribe = url(
        r'^messages/unsubscribe/(?P<message_id>\d+)/(?P<dispatch_id>\d+)/(?P<hashed>[^/]+)/$',
        unsubscribe,
        name='sitemessage_unsubscribe'
    )

    url_mark_read = url(
        r'^messages/ping/(?P<message_id>\d+)/(?P<dispatch_id>\d+)/(?P<hashed>[^/]+)/$',
        mark_read,
        name='sitemessage_mark_read'
    )

    if VERSION >= (1, 9):
        return [url_unsubscribe, url_mark_read]

    return patterns('', url_unsubscribe, url_mark_read)



from optparse import make_option

from django.core.management.base import BaseCommand

from sitemessage.toolbox import send_scheduled_messages


class Command(BaseCommand):

    help = 'Sends scheduled messages (both in pending and error statuses).'

    option_list = BaseCommand.option_list + (
        make_option('--priority', action='store', dest='priority', default=None,
                    help='Allows to filter scheduled messages by a priority number. Defaults to None.'),
    )

    def handle(self, *args, **options):
        priority = options.get('priority', None)
        priority_str = ''

        if priority is not None:
            priority_str = 'with priority %s ' % priority

        self.stdout.write('Sending scheduled messages %s...\n' % priority_str)
        try:
            send_scheduled_messages(priority=priority)
        except Exception as e:
            self.stderr.write(self.style.ERROR('Error on send: %s\n' % e))
        else:
            self.stdout.write('Sending done.\n')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings
import sitemessage.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Dispatch',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('time_created', models.DateTimeField(auto_now_add=True, verbose_name='Time created')),
                ('time_dispatched', models.DateTimeField(help_text='Time of the last delivery attempt.', verbose_name='Time dispatched', null=True, editable=False, blank=True)),
                ('messenger', models.CharField(help_text='Messenger class identifier.', max_length=250, verbose_name='Messenger', db_index=True)),
                ('address', models.CharField(help_text='Recipient address.', max_length=250, verbose_name='Address')),
                ('retry_count', models.PositiveIntegerField(default=0, help_text='A number of delivery retries has already been made.', verbose_name='Retry count')),
                ('message_cache', models.TextField(verbose_name='Message cache', null=True, editable=False)),
                ('dispatch_status', models.PositiveIntegerField(default=1, verbose_name='Dispatch status', choices=[(1, 'Pending'), (2, 'Sent'), (3, 'Error'), (4, 'Failed')])),
                ('read_status', models.PositiveIntegerField(default=0, verbose_name='Read status', choices=[(0, 'Unread'), (1, 'Read')])),
            ],
            options={
                'verbose_name': 'Dispatch',
                'verbose_name_plural': 'Dispatches',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DispatchError',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('time_created', models.DateTimeField(auto_now_add=True, verbose_name='Time created')),
                ('error_log', models.TextField(verbose_name='Text')),
                ('dispatch', models.ForeignKey(verbose_name='Dispatch', to='sitemessage.Dispatch')),
            ],
            options={
                'verbose_name': 'Dispatch error',
                'verbose_name_plural': 'Dispatch errors',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('time_created', models.DateTimeField(auto_now_add=True, verbose_name='Time created')),
                ('cls', models.CharField(help_text='Message logic class identifier.', max_length=250, verbose_name='Message class', db_index=True)),
                ('context', sitemessage.models.ContextField(verbose_name='Message context')),
                ('priority', models.PositiveIntegerField(default=0, help_text='Number describing message sending priority. Messages with different priorities can be sent with different periodicity.', verbose_name='Priority', db_index=True)),
                ('dispatches_ready', models.BooleanField(default=False, help_text='Indicates whether dispatches for this message are already formed and ready to delivery.', db_index=True, verbose_name='Dispatches ready')),
                ('sender', models.ForeignKey(verbose_name='Sender', blank=True, to=settings.AUTH_USER_MODEL, null=True)),
            ],
            options={
                'verbose_name': 'Message',
                'verbose_name_plural': 'Messages',
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='dispatch',
            name='message',
            field=models.ForeignKey(verbose_name='Message', to='sitemessage.Message'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='dispatch',
            name='recipient',
            field=models.ForeignKey(verbose_name='Recipient', blank=True, to=settings.AUTH_USER_MODEL, null=True),
            preserve_default=True,
        ),
    ]



from ..utils import register_message_types


def register_builtin_message_types():
    """Registers the built-in message types."""
    from .plain import PlainTextMessage
    from .email import EmailTextMessage, EmailHtmlMessage
    register_message_types(PlainTextMessage, EmailTextMessage, EmailHtmlMessage)



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class SitemessageConfig(AppConfig):
    """Sitemessage configuration."""

    name = 'sitemessage'
    verbose_name = _('Messaging')

    def ready(self):
        from sitemessage.utils import import_project_sitemessage_modules
        import_project_sitemessage_modules()

        from sitemessage.settings import INIT_BUILTIN_MESSAGE_TYPES
        if INIT_BUILTIN_MESSAGE_TYPES:
            from sitemessage.messages import register_builtin_message_types
            register_builtin_message_types()



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding field 'Message.priority'
        db.add_column('sitemessage_message', 'priority',
                      self.gf('django.db.models.fields.PositiveIntegerField')(default=0, db_index=True),
                      keep_default=False)


    def backwards(self, orm):
        # Deleting field 'Message.priority'
        db.delete_column('sitemessage_message', 'priority')


    models = {
        'apps.place': {
            'Meta': {'object_name': 'Place'},
            'geo_bounds': ('django.db.models.fields.CharField', [], {'null': 'True', 'max_length': '255', 'blank': 'True'}),
            'geo_pos': ('django.db.models.fields.CharField', [], {'null': 'True', 'max_length': '255', 'blank': 'True'}),
            'geo_title': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'geo_type': ('django.db.models.fields.CharField', [], {'null': 'True', 'db_index': 'True', 'max_length': '25', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'raters_num': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'rating': ('django.db.models.fields.FloatField', [], {'default': '0', 'db_index': 'True'}),
            'status': ('django.db.models.fields.PositiveIntegerField', [], {'default': '1'}),
            'time_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'time_modified': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'auto_now': 'True', 'blank': 'True'}),
            'time_published': ('django.db.models.fields.DateTimeField', [], {'null': 'True'}),
            'user_title': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'apps.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'blank': 'True', 'max_length': '75'}),
            'first_name': ('django.db.models.fields.CharField', [], {'blank': 'True', 'max_length': '30'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'blank': 'True', 'related_name': "'user_set'", 'symmetrical': 'False'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'blank': 'True', 'max_length': '30'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'place': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['apps.Place']", 'null': 'True', 'related_name': "'users'", 'blank': 'True'}),
            'raters_num': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'rating': ('django.db.models.fields.FloatField', [], {'default': '0', 'db_index': 'True'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'blank': 'True', 'related_name': "'user_set'", 'symmetrical': 'False'}),
            'username': ('django.db.models.fields.CharField', [], {'max_length': '30', 'unique': 'True'})
        },
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '80', 'unique': 'True'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'blank': 'True', 'symmetrical': 'False'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'contenttypes.contenttype': {
            'Meta': {'db_table': "'django_content_type'", 'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType'},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'sitemessage.dispatch': {
            'Meta': {'object_name': 'Dispatch'},
            'address': ('django.db.models.fields.CharField', [], {'max_length': '250'}),
            'dispatch_status': ('django.db.models.fields.PositiveIntegerField', [], {'default': '1'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'message': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['sitemessage.Message']"}),
            'message_cache': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'messenger': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '250'}),
            'read_status': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'recipient': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['apps.User']", 'null': 'True', 'blank': 'True'}),
            'retry_count': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'time_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'time_dispatched': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'})
        },
        'sitemessage.dispatcherror': {
            'Meta': {'object_name': 'DispatchError'},
            'dispatch': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['sitemessage.Dispatch']"}),
            'error_log': ('django.db.models.fields.TextField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'time_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'})
        },
        'sitemessage.message': {
            'Meta': {'object_name': 'Message'},
            'cls': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '250'}),
            'context': ('sitemessage.models.ContextField', [], {}),
            'dispatches_ready': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'db_index': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'priority': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0', 'db_index': 'True'}),
            'sender': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['apps.User']", 'null': 'True', 'blank': 'True'}),
            'time_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['sitemessage']


# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Message'
        db.create_table('sitemessage_message', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('time_created', self.gf('django.db.models.fields.DateTimeField')(blank=True, auto_now_add=True)),
            ('sender', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, to=orm['apps.User'], null=True)),
            ('cls', self.gf('django.db.models.fields.CharField')(db_index=True, max_length=250)),
            ('context', self.gf('sitemessage.models.ContextField')()),
            ('dispatches_ready', self.gf('django.db.models.fields.BooleanField')(db_index=True, default=False)),
        ))
        db.send_create_signal('sitemessage', ['Message'])

        # Adding model 'Dispatch'
        db.create_table('sitemessage_dispatch', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('time_created', self.gf('django.db.models.fields.DateTimeField')(blank=True, auto_now_add=True)),
            ('time_dispatched', self.gf('django.db.models.fields.DateTimeField')(blank=True, null=True)),
            ('message', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['sitemessage.Message'])),
            ('messenger', self.gf('django.db.models.fields.CharField')(db_index=True, max_length=250)),
            ('recipient', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, to=orm['apps.User'], null=True)),
            ('address', self.gf('django.db.models.fields.CharField')(max_length=250)),
            ('retry_count', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
            ('message_cache', self.gf('django.db.models.fields.TextField')(null=True)),
            ('dispatch_status', self.gf('django.db.models.fields.PositiveIntegerField')(default=1)),
            ('read_status', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
        ))
        db.send_create_signal('sitemessage', ['Dispatch'])

        # Adding model 'DispatchError'
        db.create_table('sitemessage_dispatcherror', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('time_created', self.gf('django.db.models.fields.DateTimeField')(blank=True, auto_now_add=True)),
            ('dispatch', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['sitemessage.Dispatch'])),
            ('error_log', self.gf('django.db.models.fields.TextField')()),
        ))
        db.send_create_signal('sitemessage', ['DispatchError'])


    def backwards(self, orm):
        # Deleting model 'Message'
        db.delete_table('sitemessage_message')

        # Deleting model 'Dispatch'
        db.delete_table('sitemessage_dispatch')

        # Deleting model 'DispatchError'
        db.delete_table('sitemessage_dispatcherror')


    models = {
        'apps.place': {
            'Meta': {'object_name': 'Place'},
            'geo_bounds': ('django.db.models.fields.CharField', [], {'blank': 'True', 'max_length': '255', 'null': 'True'}),
            'geo_pos': ('django.db.models.fields.CharField', [], {'blank': 'True', 'max_length': '255', 'null': 'True'}),
            'geo_title': ('django.db.models.fields.TextField', [], {'blank': 'True', 'null': 'True'}),
            'geo_type': ('django.db.models.fields.CharField', [], {'blank': 'True', 'db_index': 'True', 'max_length': '25', 'null': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'raters_num': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'rating': ('django.db.models.fields.FloatField', [], {'db_index': 'True', 'default': '0'}),
            'status': ('django.db.models.fields.PositiveIntegerField', [], {'default': '1'}),
            'time_created': ('django.db.models.fields.DateTimeField', [], {'blank': 'True', 'auto_now_add': 'True'}),
            'time_modified': ('django.db.models.fields.DateTimeField', [], {'blank': 'True', 'auto_now': 'True', 'null': 'True'}),
            'time_published': ('django.db.models.fields.DateTimeField', [], {'null': 'True'}),
            'user_title': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'apps.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'blank': 'True', 'max_length': '75'}),
            'first_name': ('django.db.models.fields.CharField', [], {'blank': 'True', 'max_length': '30'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'blank': 'True', 'to': "orm['auth.Group']", 'related_name': "'user_set'", 'symmetrical': 'False'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'blank': 'True', 'max_length': '30'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'place': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'to': "orm['apps.Place']", 'related_name': "'users'", 'null': 'True'}),
            'raters_num': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'rating': ('django.db.models.fields.FloatField', [], {'db_index': 'True', 'default': '0'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'blank': 'True', 'to': "orm['auth.Permission']", 'related_name': "'user_set'", 'symmetrical': 'False'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'blank': 'True', 'to': "orm['auth.Permission']", 'symmetrical': 'False'})
        },
        'auth.permission': {
            'Meta': {'object_name': 'Permission', 'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)"},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'contenttypes.contenttype': {
            'Meta': {'db_table': "'django_content_type'", 'object_name': 'ContentType', 'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'sitemessage.dispatch': {
            'Meta': {'object_name': 'Dispatch'},
            'address': ('django.db.models.fields.CharField', [], {'max_length': '250'}),
            'dispatch_status': ('django.db.models.fields.PositiveIntegerField', [], {'default': '1'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'message': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['sitemessage.Message']"}),
            'message_cache': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'messenger': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '250'}),
            'read_status': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'recipient': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'to': "orm['apps.User']", 'null': 'True'}),
            'retry_count': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'time_created': ('django.db.models.fields.DateTimeField', [], {'blank': 'True', 'auto_now_add': 'True'}),
            'time_dispatched': ('django.db.models.fields.DateTimeField', [], {'blank': 'True', 'null': 'True'})
        },
        'sitemessage.dispatcherror': {
            'Meta': {'object_name': 'DispatchError'},
            'dispatch': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['sitemessage.Dispatch']"}),
            'error_log': ('django.db.models.fields.TextField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'time_created': ('django.db.models.fields.DateTimeField', [], {'blank': 'True', 'auto_now_add': 'True'})
        },
        'sitemessage.message': {
            'Meta': {'object_name': 'Message'},
            'cls': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '250'}),
            'context': ('sitemessage.models.ContextField', [], {}),
            'dispatches_ready': ('django.db.models.fields.BooleanField', [], {'db_index': 'True', 'default': 'False'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'sender': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'to': "orm['apps.User']", 'null': 'True'}),
            'time_created': ('django.db.models.fields.DateTimeField', [], {'blank': 'True', 'auto_now_add': 'True'})
        }
    }

    complete_apps = ['sitemessage']


from .messages.email import EmailHtmlMessage, EmailTextMessage
from .toolbox import schedule_messages, recipients
from .settings import DEFAULT_SHORTCUT_EMAIL_MESSAGES_TYPE


def schedule_email(message, to, subject=None, sender=None, priority=None):
    """Schedules an email message for delivery.

    :param dict, str message: str or dict: use str for simple text email;
        dict - to compile email from a template (default: `sitemessage/messages/email_html__smtp.html`).
    :param list to: recipients addresses or Django User model heir instances
    :param str subject: email subject
    :param User sender: User model heir instance
    :param int priority: number describing message priority. If set overrides priority provided with message type.
    """
    if isinstance(message, dict):
        message_cls = EmailHtmlMessage
    else:
        message_cls = EmailTextMessage

    schedule_messages(
        message_cls(subject, message),
        recipients(DEFAULT_SHORTCUT_EMAIL_MESSAGES_TYPE, to),
        sender=sender, priority=priority
    )


def schedule_jabber_message(message, to, sender=None, priority=None):
    """Schedules Jabber XMPP message for delivery.

    :param str message: text to send.
    :param list to: recipients addresses or Django User model heir instances
    :param User sender: User model heir instance
    :param int priority: number describing message priority. If set overrides priority provided with message type.
    """
    schedule_messages(message, recipients('xmppsleek', to), sender=sender, priority=priority)


def schedule_tweet(message, to='', sender=None, priority=None):
    """Schedules a Tweet for delivery.

    :param str message: text to send.
    :param list to: recipients addresses or Django User model heir instances. If supplied tweets will be @-replies.
    :param User sender: User model heir instance
    :param int priority: number describing message priority. If set overrides priority provided with message type.
    """
    schedule_messages(message, recipients('twitter', to), sender=sender, priority=priority)


def schedule_telegram_message(message, to, sender=None, priority=None):
    """Schedules Telegram message for delivery.

    :param str message: text to send.
    :param list to: recipients addresses or Django User model heir instances
    :param User sender: User model heir instance
    :param int priority: number describing message priority. If set overrides priority provided with message type.
    """
    schedule_messages(message, recipients('telegram', to), sender=sender, priority=priority)



from django import template
from django.template.base import FilterExpression
from django.template.loader import get_template
from django.conf import settings

from ..exceptions import SiteMessageConfigurationError


register = template.Library()


@register.tag
def sitemessage_prefs_table(parser, token):
    tokens = token.split_contents()
    use_template = detect_clause(parser, 'template', tokens)
    prefs_obj = detect_clause(parser, 'from', tokens)

    tokens_num = len(tokens)

    if tokens_num in (1, 3):
        return sitemessage_prefs_tableNode(prefs_obj, use_template)
    else:
        raise template.TemplateSyntaxError(
            '`sitemessage_prefs_table` tag expects the following notation: '
            '{% sitemessage_prefs_table from user_prefs template "sitemessage/my_pref_table.html" %}.')


class sitemessage_prefs_tableNode(template.Node):

    def __init__(self, prefs_obj, use_template):
        self.use_template = use_template
        self.prefs_obj = prefs_obj

    def render(self, context):
        resolve = lambda arg: arg.resolve(context) if isinstance(arg, FilterExpression) else arg

        prefs_obj = resolve(self.prefs_obj)
        if not isinstance(prefs_obj, tuple):
            if settings.DEBUG:
                raise SiteMessageConfigurationError(
                    '`sitemessage_prefs_table` template tag expects a tuple generated '
                    'by `get_user_preferences_for_ui` but `%s` is given.' % type(prefs_obj))
            return ''  # Silent fall.

        context.push()
        context['sitemessage_user_prefs'] = prefs_obj
        contents = get_template(
            resolve(self.use_template or 'sitemessage/user_prefs_table.html')
        ).render(context)
        context.pop()

        return contents


def detect_clause(parser, clause_name, tokens):
    """Helper function detects a certain clause in tag tokens list.
    Returns its value.

    """
    if clause_name in tokens:
        t_index = tokens.index(clause_name)
        clause_value = parser.compile_filter(tokens[t_index + 1])
        del tokens[t_index:t_index + 2]
    else:
        clause_value = None
    return clause_value



from django.utils.translation import ugettext as _

from .base import MessageBase


class _EmailMessageBase(MessageBase):

    supported_messengers = ['smtp']
    title = _('Email notification')

    def __init__(self, subject, text_or_dict, type_name, template_path=None):
        context = {
            'subject': subject,
            'type': type_name,
        }
        self.update_context(context, text_or_dict)
        super(_EmailMessageBase, self).__init__(context, template_path=template_path)


class EmailTextMessage(_EmailMessageBase):
    """Simple plain text message to send as an e-mail."""

    alias = 'email_plain'
    template_ext = 'txt'

    def __init__(self, subject, text_or_dict, template_path=None):
        super(EmailTextMessage, self).__init__(subject, text_or_dict, 'plain', template_path=template_path)


class EmailHtmlMessage(_EmailMessageBase):
    """HTML message to send as an e-mail."""

    alias = 'email_html'
    template_ext = 'html'

    def __init__(self, subject, html_or_dict, template_path=None):
        super(EmailHtmlMessage, self).__init__(subject, html_or_dict, 'html', template_path=template_path)



"""Compatibility module with misspelled name.

"""
import warnings

from .shortcuts import *

warnings.warn(
    'Import from `schortcuts` module is deprecated. '
    'Will be removed in 1.0. Please import from `shortcuts`.', DeprecationWarning)



from mock import MagicMock, patch
from django.core.mail import send_mail
from django.core.urlresolvers import reverse
from django.core.management import call_command
from django.test.utils import override_settings
from django.test import TestCase
from django.test.client import RequestFactory, Client
from django.contrib.auth.models import User

try:
    from django.template.base import TemplateDoesNotExist, Template, TemplateSyntaxError
except ImportError:
    # Django 1.9+
    from django.template import TemplateDoesNotExist, Template, TemplateSyntaxError

from django.template.context import Context

from .messages.plain import PlainTextMessage
from .messages.base import MessageBase
from .models import Message, Dispatch, Subscription, DispatchError
from .toolbox import schedule_messages, recipients, send_scheduled_messages, prepare_dispatches, \
    get_user_preferences_for_ui, register_builtin_message_types, get_sitemessage_urls, \
    set_user_preferences_from_request, _ALIAS_SEP, _PREF_POST_KEY
from .messengers.base import MessengerBase
from .utils import Recipient, register_messenger_objects, \
    register_message_types, get_registered_messenger_objects, get_registered_messenger_object, \
    get_registered_message_types, override_message_type_for_app, get_message_type_for_app
from .exceptions import UnknownMessengerError, SiteMessageConfigurationError
from .signals import sig_mark_read_failed, sig_mark_read_success, sig_unsubscribe_failed, sig_unsubscribe_success

from .shortcuts import schedule_email, schedule_jabber_message
from .messengers.smtp import SMTPMessenger
from .messengers.xmpp import XMPPSleekMessenger
from .messengers.twitter import TwitterMessenger
from .messengers.telegram import TelegramMessenger


WONDERLAND_DOMAIN = '@wonderland'

urlpatterns = get_sitemessage_urls()


class MockException(Exception):
    """This will prevent `catching classes that do not inherit from BaseException is not allowed` errors
    when `mock_thirdparty` is used.

    """


def mock_thirdparty(name, func, mock=None):
    if mock is None:
        mock = MagicMock()

    with patch.dict('sys.modules', {name: mock}):
        result = func()

    return result


messenger_smtp = mock_thirdparty('smtplib', lambda: SMTPMessenger(login='someone', use_tls=True))
messenger_xmpp = mock_thirdparty('sleekxmpp', lambda: XMPPSleekMessenger('somjid', 'somepasswd'))
messenger_xmpp._session_started = True
messenger_twitter = mock_thirdparty('twitter', lambda: TwitterMessenger('key', 'secret', 'token', 'token_secret'))
messenger_twitter.lib = MagicMock()
messenger_telegram = mock_thirdparty('requests', lambda: TelegramMessenger('bottoken'))
messenger_telegram.lib = MagicMock()
messenger_telegram.lib.exceptions.RequestException = MockException

register_messenger_objects(messenger_smtp, messenger_xmpp, messenger_twitter, messenger_telegram)

register_builtin_message_types()


class TestMessenger(MessengerBase):

    title = 'Test messenger'
    alias = 'test_messenger'
    last_send = None

    def __init__(self, login, password):
        self.login = login
        self.password = password

    @classmethod
    def get_address(cls, recipient):
        if isinstance(recipient, User):
            recipient = 'user_%s' % recipient.username

        return '%s%s' % (recipient, WONDERLAND_DOMAIN)

    def send(self, message_cls, message_model, dispatch_models):
        self.last_send = {
            'message_cls': message_cls,
            'dispatch_models': dispatch_models,
            'message_model': message_model,
        }


class BuggyMessenger(MessengerBase):

    title = 'Buggy messenger'
    alias = 'buggy'

    def send(self, message_cls, message_model, dispatch_models):
        raise Exception('Damn it.')


class TestMessage(MessageBase):

    alias = 'test_message'
    template_ext = 'html'
    title = 'Test message type'
    supported_messengers = ['smtp', 'test_messenger']


class TestMessagePlain(PlainTextMessage):

    alias = 'testplain'
    priority = 10


class TestMessagePlainDynamic(PlainTextMessage):

    alias = 'testplain_dyn'
    has_dynamic_context = True

    @classmethod
    def compile(cls, message, messenger, dispatch=None):
        return '%s -- %s' % (message.context[MessageBase.SIMPLE_TEXT_ID], dispatch.address)


register_messenger_objects(TestMessenger('mylogin', 'mypassword'), BuggyMessenger())
register_message_types(PlainTextMessage, TestMessage, TestMessagePlain, TestMessagePlainDynamic)


class SitemessageTest(TestCase):

    def assert_called_n(self, func, n=1):
        self.assertEqual(func.call_count, n)
        func.call_count = 0

    def tearDown(self):
        User.objects.all().delete()
        Message.objects.all().delete()
        Dispatch.objects.all().delete()
        DispatchError.objects.all().delete()
        Subscription.objects.all().delete()

    @classmethod
    def render_template(cls, string, context_dict=None):
        context_dict = context_dict or {}
        return Template(string).render(Context(context_dict))


class ToolboxTest(SitemessageTest):

    def test_get_user_preferences_for_ui(self):

        user = User()
        user.save()

        messengers_titles, prefs = get_user_preferences_for_ui(user)

        self.assertEqual(len(prefs.keys()), 3)
        self.assertEqual(len(messengers_titles), 6)

        Subscription.create(user, TestMessage, TestMessenger)
        user_prefs = get_user_preferences_for_ui(
            user,
            message_filter=lambda m: m.alias == 'test_message',
            messenger_filter=lambda m: m.alias in ['smtp', 'test_messenger']
        )
        messengers_titles, prefs = user_prefs

        self.assertEqual(len(prefs.keys()), 1)
        self.assertEqual(len(messengers_titles), 2)
        self.assertIn('E-mail', messengers_titles)
        self.assertIn('Test messenger', messengers_titles)

        html = self.render_template(
            "{% load sitemessage %}{% sitemessage_prefs_table from user_prefs %}",
            {'user_prefs': user_prefs}
        )

        self.assertIn('class="sitemessage_prefs', html)
        self.assertIn('E-mail</th>', html)
        self.assertIn('Test messenger</th>', html)
        self.assertIn('Test messenger</th>', html)
        self.assertIn('value="test_message|smtp"', html)
        self.assertIn('value="test_message|test_messenger" checked', html)

        prefs_row = prefs.popitem()
        self.assertEqual(prefs_row[0], 'Test message type')
        self.assertIn(('test_message|smtp', True, False), prefs_row[1])
        self.assertIn(('test_message|test_messenger', True, True), prefs_row[1])

    def test_templatetag_fails_silent(self):
        html = self.render_template(
            "{% load sitemessage %}{% sitemessage_prefs_table from user_prefs %}",
            {'user_prefs': 'a'}
        )
        self.assertEqual(html, '')

    @override_settings(DEBUG=True)
    def test_templatetag_fails_loud(self):
        tpl = "{% load sitemessage %}{% sitemessage_prefs_table from user_prefs %}"
        context = {'user_prefs': 'a'}
        self.assertRaises(SiteMessageConfigurationError, self.render_template, tpl, context)

        tpl = "{% load sitemessage %}{% sitemessage_prefs_table user_prefs %}"
        self.assertRaises(TemplateSyntaxError, self.render_template, tpl)

    def test_send_scheduled_messages_unknown_messenger(self):
        message = Message()
        message.save()
        dispatch = Dispatch(message=message, messenger='unknownname')
        dispatch.save()

        self.assertRaises(UnknownMessengerError, send_scheduled_messages)

        send_scheduled_messages(ignore_unknown_messengers=True)

    def test_set_user_preferences_from_request(self):
        user = User()
        user.save()

        r = RequestFactory().post('/', data={_PREF_POST_KEY: 'aaa%sqqq' % _ALIAS_SEP})
        r.user = user
        set_user_preferences_from_request(r)

        subs = Subscription.objects.all()
        self.assertEqual(len(subs), 0)

        r = RequestFactory().post('/', data={_PREF_POST_KEY: 'test_message%stest_messenger' % _ALIAS_SEP})
        r.user = user
        set_user_preferences_from_request(r)

        subs = Subscription.objects.all()
        self.assertEqual(len(subs), 1)


class UtilsTest(SitemessageTest):

    def test_register_messengers(self):
        messenger = type('MyMessenger', (MessengerBase,), {})
        register_messenger_objects(messenger)
        self.assertIn(messenger.get_alias(), get_registered_messenger_objects())

    def test_register_message_types(self):
        message = type('MyMessage', (MessageBase,), {})
        register_message_types(message)
        self.assertIn(message.get_alias(), get_registered_message_types())

    def test_recipients(self):
        user = User(username='myuser')
        to = ('gogi', 'givi', user)

        r1 = recipients('test_messenger', to)

        self.assertEqual(len(r1), len(to))
        self.assertEqual(r1[0].address, 'gogi%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[0].messenger, 'test_messenger')
        self.assertEqual(r1[1].address, 'givi%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[1].messenger, 'test_messenger')
        self.assertEqual(r1[2].address, 'user_myuser%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[2].messenger, 'test_messenger')

    def test_prepare_undispatched(self):
        m, d = Message.create('testplain', {MessageBase.SIMPLE_TEXT_ID: 'abc'})

        Subscription.create('fred', 'testplain', 'test_messenger')
        Subscription.create('colon', 'testplain', 'test_messenger')

        dispatches = prepare_dispatches()
        self.assertEqual(len(dispatches), 2)
        self.assertEqual(dispatches[0].address, 'fred')
        self.assertEqual(dispatches[1].address, 'colon')

    def test_send_scheduled_messages(self):
        # This one won't count, as won't fit into message priority filter.
        schedule_messages(TestMessagePlainDynamic('my_dyn_msg'), recipients('test_messenger', ('three', 'four')))

        msgr = get_registered_messenger_object('test_messenger')
        msg = TestMessagePlain('my_message')
        schedule_messages(msg, recipients(msgr, ('one', 'two')))
        send_scheduled_messages(priority=TestMessagePlain.priority)
        self.assertEqual(len(msgr.last_send['dispatch_models']), 2)
        self.assertEqual(msgr.last_send['message_model'].cls, 'testplain')
        self.assertEqual(msgr.last_send['message_cls'], TestMessagePlain)
        self.assertEqual(msgr.last_send['dispatch_models'][0].message_cache, 'my_message')
        self.assertEqual(msgr.last_send['dispatch_models'][1].message_cache, 'my_message')

    def test_send_scheduled_messages_dynamic_context(self):
        msgr = get_registered_messenger_object('test_messenger')
        msg_dyn = TestMessagePlainDynamic('my_dyn_msg')
        schedule_messages(msg_dyn, recipients(msgr, ('three', 'four')))
        send_scheduled_messages()
        self.assertEqual(len(msgr.last_send['dispatch_models']), 2)
        self.assertEqual(msgr.last_send['message_model'].cls, 'testplain_dyn')
        self.assertEqual(msgr.last_send['message_cls'], TestMessagePlainDynamic)
        self.assertEqual(msgr.last_send['dispatch_models'][0].message_cache, 'my_dyn_msg -- three%s' % WONDERLAND_DOMAIN)
        self.assertEqual(msgr.last_send['dispatch_models'][1].message_cache, 'my_dyn_msg -- four%s' % WONDERLAND_DOMAIN)

    def test_schedule_message(self):
        msg = TestMessagePlain('schedule_func')
        model, _ = schedule_messages(msg)[0]
        self.assertEqual(model.cls, msg.get_alias())
        self.assertEqual(model.context, msg.get_context())
        self.assertEqual(model.priority, TestMessagePlain.priority)
        self.assertFalse(model.dispatches_ready)

        msg = TestMessagePlain('schedule_func')
        model, _ = schedule_messages(msg, priority=33)[0]
        self.assertEqual(model.cls, msg.get_alias())
        self.assertEqual(model.context, msg.get_context())
        self.assertEqual(model.priority, 33)
        self.assertFalse(model.dispatches_ready)

        user = User()
        user.save()

        model, dispatch_models = schedule_messages('simple message', recipients('test_messenger', ('gogi', 'givi')), sender=user)[0]

        self.assertEqual(model.cls, 'plain')
        self.assertEqual(model.context, {'use_tpl': False, MessageBase.SIMPLE_TEXT_ID: 'simple message', 'tpl': None})
        self.assertEqual(model.sender, user)
        self.assertTrue(model.dispatches_ready)

        self.assertEqual(len(dispatch_models), 2)
        self.assertEqual(dispatch_models[0].address, 'gogi%s' % WONDERLAND_DOMAIN)
        self.assertEqual(dispatch_models[0].messenger, 'test_messenger')

    def test_override_message_type_for_app(self):

        mt = get_message_type_for_app('myapp', 'testplain')
        self.assertIs(mt, TestMessagePlain)

        override_message_type_for_app('myapp', 'sometype', 'test_message')
        mt = get_message_type_for_app('myapp', 'sometype')
        self.assertIs(mt, TestMessage)


class SubscriptionModelTest(SitemessageTest):

    def test_create(self):

        s = Subscription.create('abc', 'message', 'messenger')

        self.assertIsNotNone(s.time_created)
        self.assertIsNone(s.recipient)
        self.assertEqual(s.address, 'abc')
        self.assertEqual(s.message_cls, 'message')
        self.assertEqual(s.messenger_cls, 'messenger')

        s = Subscription.create(1, 'message', 'messenger')

        self.assertIsNotNone(s.time_created)
        self.assertIsNone(s.address)
        self.assertEqual(s.recipient_id, 1)
        self.assertEqual(s.message_cls, 'message')
        self.assertEqual(s.messenger_cls, 'messenger')

    def test_cancel(self):

        Subscription.create('abc', 'message', 'messenger')
        Subscription.create('abc', 'message1', 'messenger')
        Subscription.create('abc', 'message', 'messenger1')
        self.assertEqual(
            Subscription.objects.filter(address='abc').count(), 3
        )

        Subscription.cancel('abc', 'message', 'messenger')
        self.assertEqual(
            Subscription.objects.filter(address='abc').count(), 2
        )

        Subscription.create(1, 'message', 'messenger')
        self.assertEqual(
            Subscription.objects.filter(recipient=1).count(), 1
        )

        Subscription.cancel(1, 'message', 'messenger')
        self.assertEqual(
            Subscription.objects.filter(recipient=1).count(), 0
        )

    def test_replace_for_user(self):
        new_prefs = [('message3', 'messenger3')]
        user = User()

        r = Subscription.replace_for_user(user, new_prefs)

        user.save()

        Subscription.create(user, 'message', 'messenger')
        Subscription.create(user, 'message2', 'messenger2')

        self.assertEqual(Subscription.get_for_user(user).count(), 2)

        Subscription.replace_for_user(user, new_prefs)

        s = Subscription.get_for_user(user)
        self.assertEqual(s.count(), 1)
        s = s[0]
        self.assertEqual(s.message_cls, 'message3')
        self.assertEqual(s.messenger_cls, 'messenger3')

    def test_get_for_user(self):
        user = User()

        r = Subscription.get_for_user(user)
        self.assertEqual(r, [])

        user.save()

        self.assertEqual(Subscription.get_for_user(user).count(), 0)

        Subscription.create(user, 'message', 'messenger')

        self.assertEqual(Subscription.get_for_user(user).count(), 1)

    def test_get_for_message_cls(self):
        self.assertEqual(Subscription.get_for_message_cls('mymsg').count(), 0)

        Subscription.create('aaa', 'mymsg', 'messenger')
        Subscription.create('bbb', 'mymsg', 'messenger2')

        self.assertEqual(Subscription.get_for_message_cls('mymsg').count(), 2)

    def test_str(self):
        s = Subscription()
        s.address = 'aaa'

        self.assertIn('aaa', str(s))


class DispatchErrorModelTest(SitemessageTest):

    def test_str(self):
        e = DispatchError()
        e.dispatch_id = 444

        self.assertIn('444', str(e))


class DispatchModelTest(SitemessageTest):

    def test_create(self):

        message = Message(cls='test_message')
        message.save()

        user = User(username='u')
        user.save()

        recipients_ = recipients('test_messenger', [user])
        recipients_ += recipients('buggy', 'idle')

        dispatches = Dispatch.create(message, recipients_)
        self.assertTrue(isinstance(dispatches[0], Dispatch))
        self.assertTrue(isinstance(dispatches[1], Dispatch))
        self.assertEqual(dispatches[0].message_id, message.id)
        self.assertEqual(dispatches[1].message_id, message.id)
        self.assertEqual(dispatches[0].messenger, 'test_messenger')
        self.assertEqual(dispatches[1].messenger, 'buggy')

        dispatches = Dispatch.create(message, Recipient('msgr', None, 'address'))
        self.assertEqual(len(dispatches), 1)

    def test_log_dispatches_errors(self):

        self.assertEqual(DispatchError.objects.count(), 0)

        d1 = Dispatch(message_id=1)
        d1.save()

        d1.error_log = 'some_text'

        Dispatch.log_dispatches_errors([d1])
        self.assertEqual(DispatchError.objects.count(), 1)
        self.assertEqual(DispatchError.objects.get(pk=1).error_log, 'some_text')

    def test_get_unread(self):

        d1 = Dispatch(message_id=1)
        d1.save()

        d2 = Dispatch(message_id=1)
        d2.save()
        self.assertEqual(Dispatch.get_unread().count(), 2)

        d2.read_status = Dispatch.READ_STATUS_READ
        d2.save()
        self.assertEqual(Dispatch.get_unread().count(), 1)

    def test_set_dispatches_statuses(self):

        d = Dispatch(message_id=1)
        d.save()

        Dispatch.set_dispatches_statuses(**{'sent': [d]})
        d_ = Dispatch.objects.get(pk=d.id)
        self.assertEqual(d_.dispatch_status, Dispatch.DISPATCH_STATUS_SENT)
        self.assertEqual(d_.retry_count, 1)

        Dispatch.set_dispatches_statuses(**{'error': [d]})
        d_ = Dispatch.objects.get(pk=d.id)
        self.assertEqual(d_.dispatch_status, Dispatch.DISPATCH_STATUS_ERROR)
        self.assertEqual(d_.retry_count, 2)

    def test_str(self):
        d = Dispatch()
        d.address = 'tttt'

        self.assertIn('tttt', str(d))

    def test_mark_read(self):
        d = Dispatch()
        self.assertEqual(d.read_status, d.READ_STATUS_UNDREAD)
        d.mark_read()
        self.assertEqual(d.read_status, d.READ_STATUS_READ)


class MessageModelTest(SitemessageTest):

    def test_create(self):
        user = User(username='u')
        user.save()

        m, _ = Message.create('some', {'abc': 'abc'}, sender=user, priority=22)
        self.assertEqual(m.cls, 'some')
        self.assertEqual(m.context, {'abc': 'abc'})
        self.assertEqual(m.sender, user)
        self.assertEqual(m.priority, 22)
        self.assertFalse(m.dispatches_ready)

        ctx = {'a': 'a', 'b': 'b'}
        m, _ = Message.create('some2', ctx)
        self.assertEqual(m.cls, 'some2')
        self.assertEqual(m.context, ctx)
        self.assertIsNone(m.sender)
        self.assertFalse(m.dispatches_ready)

    def test_deserialize_context(self):
        m = Message(cls='some_cls', context={'a': 'a', 'b': 'b', 'c': 'c'})
        m.save()

        m2 = Message.objects.get(pk=m.pk)
        self.assertEquals(m2.context, {'a': 'a', 'b': 'b', 'c': 'c'})

    def test_get_type(self):
        m = Message(cls='test_message')

        self.assertIs(m.get_type(), TestMessage)

    def test_str(self):
        m = Message()
        m.cls = 'aaa'

        self.assertEqual(str(m), 'aaa')


class MessengerTest(SitemessageTest):

    def test_init_params(self):
        messengers = get_registered_messenger_objects()
        my = messengers['test_messenger']
        self.assertEqual(my.login, 'mylogin')
        self.assertEqual(my.password, 'mypassword')

    def test_alias(self):
        messenger = type('MyMessenger', (MessengerBase,), {'alias': 'myalias'})
        self.assertEqual(messenger.get_alias(), 'myalias')

        messenger = type('MyMessenger', (MessengerBase,), {})
        self.assertEqual(messenger.get_alias(), 'MyMessenger')

    def test_get_recipients_data(self):
        user = User(username='myuser')
        to = ('gogi', 'givi', user)

        r1 = TestMessenger._structure_recipients_data(to)

        self.assertEqual(len(r1), len(to))
        self.assertEqual(r1[0].address, 'gogi%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[0].messenger, 'test_messenger')
        self.assertEqual(r1[1].address, 'givi%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[1].messenger, 'test_messenger')
        self.assertEqual(r1[2].address, 'user_myuser%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[2].messenger, 'test_messenger')

    def test_send(self):
        m = TestMessenger('l', 'p')
        m.send('message_cls', 'message_model', 'dispatch_models')

        self.assertEqual(m.last_send['message_cls'], 'message_cls')
        self.assertEqual(m.last_send['message_model'], 'message_model')
        self.assertEqual(m.last_send['dispatch_models'], 'dispatch_models')

        m = BuggyMessenger()
        recipiets_ = recipients('test_messenger', ['a', 'b', 'c', 'd'])
        self.assertRaises(Exception, m.send, 'a buggy message', recipiets_)


class ShortcutsTest(SitemessageTest):

    def test_schedule_email(self):
        schedule_email('some text', 'some@one.here')

        self.assertEqual(Message.objects.all()[0].cls, 'email_plain')
        self.assertEqual(Message.objects.count(), 1)
        self.assertEqual(Dispatch.objects.count(), 1)

        schedule_email({'header': 'one', 'body': 'two'}, 'some@one.here')

        self.assertEqual(Message.objects.all()[1].cls, 'email_html')
        self.assertEqual(Message.objects.count(), 2)
        self.assertEqual(Dispatch.objects.count(), 2)

    def test_schedule_jabber_message(self):
        schedule_jabber_message('message', 'noone')


class SMTPMessengerTest(SitemessageTest):

    def test_get_address(self):
        r = object()
        self.assertEqual(messenger_smtp.get_address(r), r)

        r = type('r', (object,), dict(email='somewhere'))
        self.assertEqual(messenger_smtp.get_address(r), 'somewhere')

    def test_send(self):
        schedule_messages('text', recipients('smtp', 'someone'))
        send_scheduled_messages()
        self.assert_called_n(messenger_smtp.smtp.sendmail)

    def test_send_fail(self):
        schedule_messages('text', recipients('smtp', 'someone'))

        def new_method(*args, **kwargs):
            raise Exception('smtp failed')

        old_method = messenger_smtp.smtp.sendmail
        messenger_smtp.smtp.sendmail = new_method

        try:
            send_scheduled_messages()
            errors = DispatchError.objects.all()
            self.assertEqual(len(errors), 1)
            self.assertEqual(errors[0].error_log, 'smtp failed')
            self.assertEqual(errors[0].dispatch.address, 'someone')
        finally:
            messenger_smtp.smtp.sendmail = old_method

    def test_send_test_message(self):
        messenger_smtp.send_test_message('someone', 'sometext')
        self.assert_called_n(messenger_smtp.smtp.sendmail)


class TwitterMessengerTest(SitemessageTest):

    def test_get_address(self):
        r = object()
        self.assertEqual(messenger_twitter.get_address(r), r)

        r = type('r', (object,), dict(twitter='somewhere'))
        self.assertEqual(messenger_twitter.get_address(r), 'somewhere')

    def test_send(self):
        schedule_messages('text', recipients('twitter', 'someone'))
        send_scheduled_messages()
        messenger_twitter.api.statuses.update.assert_called_with(status='@someone text')

    def test_send_test_message(self):
        messenger_twitter.send_test_message('someone', 'sometext')
        messenger_twitter.api.statuses.update.assert_called_with(status='@someone sometext')

        messenger_twitter.send_test_message('', 'sometext')
        messenger_twitter.api.statuses.update.assert_called_with(status='sometext')

    def test_send_fail(self):
        schedule_messages('text', recipients('twitter', 'someone'))

        def new_method(*args, **kwargs):
            raise Exception('tweet failed')

        old_method = messenger_twitter.api.statuses.update
        messenger_twitter.api.statuses.update = new_method

        try:
            send_scheduled_messages()
            errors = DispatchError.objects.all()
            self.assertEqual(len(errors), 1)
            self.assertEqual(errors[0].error_log, 'tweet failed')
            self.assertEqual(errors[0].dispatch.address, 'someone')
        finally:
            messenger_twitter.api.statuses.update = old_method


class XMPPSleekMessengerTest(SitemessageTest):

    def test_get_address(self):
        r = object()
        self.assertEqual(messenger_xmpp.get_address(r), r)

        r = type('r', (object,), dict(jabber='somewhere'))
        self.assertEqual(messenger_xmpp.get_address(r), 'somewhere')

    def test_send(self):
        schedule_messages('text', recipients('xmppsleek', 'someone'))
        send_scheduled_messages()
        messenger_xmpp.xmpp.send_message.assert_called_once_with(
            mtype='chat', mbody='text', mfrom='somjid', mto='someone'
        )

    def test_send_test_message(self):
        messenger_xmpp.send_test_message('someone', 'sometext')
        messenger_xmpp.xmpp.send_message.assert_called_with(
            mtype='chat', mbody='sometext', mfrom='somjid', mto='someone'
        )

    def test_send_fail(self):
        schedule_messages('text', recipients('xmppsleek', 'someone'))

        def new_method(*args, **kwargs):
            raise Exception('xmppsleek failed')

        old_method = messenger_xmpp.xmpp.send_message
        messenger_xmpp.xmpp.send_message = new_method
        messenger_xmpp._session_started = True
        try:
            send_scheduled_messages()
            errors = DispatchError.objects.all()
            self.assertEqual(len(errors), 1)
            self.assertEqual(errors[0].error_log, 'xmppsleek failed')
            self.assertEqual(errors[0].dispatch.address, 'someone')
        finally:
            messenger_xmpp.xmpp.send_message = old_method


class TelegramMessengerTest(SitemessageTest):

    def setUp(self):
        messenger_telegram._verify_bot()
        messenger_telegram.lib.post.call_count = 0

    def test_get_address(self):
        r = object()
        self.assertEqual(messenger_telegram.get_address(r), r)

        r = type('r', (object,), dict(telegram='chat_id'))
        self.assertEqual(messenger_telegram.get_address(r), 'chat_id')

    def test_send(self):
        schedule_messages('text', recipients('telegram', '1234567'))
        send_scheduled_messages()
        self.assert_called_n(messenger_telegram.lib.post)

    def test_send_test_message(self):
        messenger_telegram.send_test_message('someone', 'sometext')
        self.assert_called_n(messenger_telegram.lib.post)

        messenger_telegram.send_test_message('', 'sometext')
        self.assert_called_n(messenger_telegram.lib.post)

    def test_get_chat_ids(self):
        self.assertEqual(messenger_telegram.get_chat_ids(), [])
        self.assert_called_n(messenger_telegram.lib.post, 2)

    def test_send_fail(self):
        schedule_messages('text', recipients('telegram', 'someone'))

        def new_method(*args, **kwargs):
            raise Exception('telegram failed')

        old_method = messenger_telegram.lib.post
        messenger_telegram.lib.post = new_method

        try:
            send_scheduled_messages()
            errors = DispatchError.objects.all()
            self.assertEqual(len(errors), 1)
            self.assertEqual(errors[0].error_log, 'telegram failed')
            self.assertEqual(errors[0].dispatch.address, 'someone')
        finally:
            messenger_telegram.lib.post = old_method

class ViewsTest(SitemessageTest):

    STATUS_SUCCESS = 'success'
    STATUS_FAIL = 'fail'

    def setUp(self):

        def catcher_success(*args, **kwargs):
            self.status = self.STATUS_SUCCESS
        self.catcher_success = catcher_success

        def catcher_fail(*args, **kwargs):
            self.status = self.STATUS_FAIL
        self.catcher_fail = catcher_fail

        user = User()
        user.save()
        self.user = user

        msg_type = TestMessagePlain('sometext')
        self.msg_type = msg_type

        msg_model, _ = schedule_messages(msg_type, recipients(TestMessenger, user))[0]
        dispatch = Dispatch.objects.all()[0]
        self.msg_model = msg_model
        self.dispatch = dispatch

        dispatch_hash = msg_type.get_dispatch_hash(dispatch.id, msg_model.id)
        self.dispatch_hash = dispatch_hash

    def send_request(self, msg_id, dispatch_id, dispatch_hash, expected_status):
        Client().get(reverse(self.view_name, args=[msg_id, dispatch_id, dispatch_hash]))
        self.assertEqual(self.status, expected_status)
        self.status = None

    def generic_view_test(self):
        # Unknown dispatch ID.
        self.send_request(self.msg_model.id, 999999, self.dispatch_hash, self.STATUS_FAIL)
        # Invalid hash.
        self.send_request(self.msg_model.id, self.dispatch.id, 'nothash', self.STATUS_FAIL)
        # Message ID mismatch.
        self.send_request(999999, self.dispatch.id, self.dispatch_hash, self.STATUS_FAIL)

    def test_unsubscribe(self):
        self.view_name = 'sitemessage_unsubscribe'

        sig_unsubscribe_success.connect(self.catcher_success, weak=False)
        sig_unsubscribe_failed.connect(self.catcher_fail, weak=False)

        self.generic_view_test()

        subscr = Subscription(
            message_cls=self.msg_type, messenger_cls=TestMessenger.alias, recipient=self.user
        )
        subscr.save()
        self.assertEqual(len(Subscription.objects.all()), 1)

        self.send_request(self.msg_model.id, self.dispatch.id, self.dispatch_hash, self.STATUS_SUCCESS)
        self.assertEqual(len(Subscription.objects.all()), 0)

    def test_mark_read(self):

        self.view_name = 'sitemessage_mark_read'

        sig_mark_read_success.connect(self.catcher_success, weak=False)
        sig_mark_read_failed.connect(self.catcher_fail, weak=False)

        self.generic_view_test()
        self.assertFalse(Dispatch.objects.get(pk=self.dispatch.pk).is_read())

        self.send_request(self.msg_model.id, self.dispatch.id, self.dispatch_hash, self.STATUS_SUCCESS)
        self.assertTrue(Dispatch.objects.get(pk=self.dispatch.pk).is_read())


class MessageTest(SitemessageTest):

    def test_alias(self):
        message = type('MyMessage', (MessageBase,), {'alias': 'myalias'})
        self.assertEqual(message.get_alias(), 'myalias')

        message = type('MyMessage', (MessageBase,), {})
        self.assertEqual(message.get_alias(), 'MyMessage')

        message = message()
        self.assertEqual(str(message), 'MyMessage')

    def test_context(self):
        msg = TestMessage({'title': 'My message!', 'name': 'idle'})
        self.assertEqual(msg.context, {'name': 'idle', 'title': 'My message!', 'tpl': None, 'use_tpl': True})

    def test_get_template(self):
        self.assertEqual(
            TestMessage.get_template(TestMessage(), TestMessenger('a', 'b')),
            'sitemessage/messages/test_message__test_messenger.html'
        )

    def test_compile_string(self):
        msg = TestMessagePlain('simple')
        self.assertEqual(TestMessagePlain.compile(msg, TestMessenger('a', 'b')), 'simple')

    def test_compile_dict(self):
        msg = TestMessage({'title': 'My message!', 'name': 'idle'})
        self.assertRaises(TemplateDoesNotExist, TestMessage.compile, msg, TestMessenger('a', 'b'))

    def test_schedule(self):
        msg = TestMessagePlain('schedule1')
        model, _ = msg.schedule()
        self.assertEqual(model.cls, msg.get_alias())
        self.assertEqual(model.context, msg.get_context())
        self.assertIsNone(model.sender)

        user = User()
        user.save()

        msg = TestMessagePlain('schedule2')
        model, _ = msg.schedule(sender=user)
        self.assertEqual(model.sender, user)


class CommandsTest(SitemessageTest):

    def test_send_scheduled(self):
        call_command('sitemessage_send_scheduled', priority=1)


@override_settings(EMAIL_BACKEND='sitemessage.backends.EmailBackend')
class BackendsTest(SitemessageTest):

    def test_email_backend(self):
        send_mail('subj', 'message', 'from@example.com', ['to@example.com'], fail_silently=False)
        dispatches = list(Dispatch.objects.all())
        self.assertEqual(len(dispatches), 1)
        self.assertEqual(dispatches[0].messenger, 'smtp')
        self.assertEqual(dispatches[0].address, 'to@example.com')
        self.assertEqual(dispatches[0].dispatch_status, Dispatch.DISPATCH_STATUS_PENDING)
        self.assertEqual(dispatches[0].message.cls, 'email_html')
        self.assertEqual(dispatches[0].message.context['subject'], 'subj')
        self.assertEqual(dispatches[0].message.context['contents'], 'message')



from django.core.urlresolvers import reverse, NoReverseMatch
from django.utils.crypto import salted_hmac
from django.utils.translation import ugettext as _
from django.shortcuts import redirect
from django.template.loader import render_to_string
from django.utils.six import string_types

from ..utils import Recipient, recipients, get_site_url, get_registered_messenger_object
from ..models import Message, Dispatch, Subscription
from ..signals import sig_unsubscribe_success, sig_unsubscribe_failed, sig_mark_read_success, sig_mark_read_failed
from ..exceptions import UnknownMessengerError


APP_URLS_ATTACHED = None


class MessageBase(object):
    """Base class for messages used by sitemessage.
    Customized message handling is available through inheritance.

    """

    # List of supported messengers (aliases).
    supported_messengers = []

    # Message type alias to address it from different places, Should rather be quite unique %)
    alias = None

    # Title to show to user.
    title = _('Notification')

    # Number describing message priority. Can be overridden by `priority` provided with schedule_messages().
    priority = None

    # This flag is used to optimize template compilation process.
    # If True template will be compiled for every dispatch (and dispatch data will be available in it)
    # instead of just once per message.
    has_dynamic_context = False

    # Template file extension. Considered when the below mentioned `template` field is not set.
    template_ext = 'tpl'

    # Path to the template to be used for message rendering.
    # If not set, will be deduced from message, messenger data (e.g. `sitemessage/plain_smtp.txt`)
    # and `template_ext` (see above).
    template = None

    # This limits the number of send attempts before message delivery considered failed.
    send_retry_limit = 10

    # Makes subscription for this message type available for users (see get_user_preferences_for_ui())
    allow_user_subscription = True

    _message_model = None
    _dispatch_models = None

    SIMPLE_TEXT_ID = 'stext_'

    def __init__(self, context=None, template_path=None):
        """Initializes a message.

        :param dict, str context: data to be used for message rendering (e.g. in templates)
        :param str template_path: template path
        """
        context_base = {
            'tpl': None,  # Template path to use
            'use_tpl': False  # Use template boolean flag
        }

        if context is not None:
            self.update_context(context_base, context, template_path=template_path)

        self.context = context_base

    @classmethod
    def get_alias(cls):
        """Returns message type alias.

        :return: str
        :rtype: str
        """
        if cls.alias is None:
            cls.alias = cls.__name__
        return cls.alias

    def __str__(self):
        return self.__class__.get_alias()

    def get_context(self):
        """Returns message context.

        :return: dict
        :rtype: dict
        """
        return self.context

    def schedule(self, recipients=None, sender=None, priority=None):
        """Schedules message for a delivery.
        Puts message (and dispatches if any) data into DB.

        :param list|None recipients: recipient (or a list) or None.
            If `None` Dispatches should be created before send using `prepare_dispatches()`.
        :param User|None sender: Django User model heir instance
        :param int|None priority: number describing message priority
        :return: a tuple with message model and a list of dispatch models.
        :rtype: tuple
        """
        if priority is None:
            priority = self.priority

        self._message_model, self._dispatch_models = Message.create(
            self.get_alias(), self.get_context(), recipients=recipients, sender=sender, priority=priority
        )
        return self._message_model, self._dispatch_models

    @classmethod
    def recipients(cls, messenger, addresses):
        """Shorcut method. See `recipients()`,"""
        return recipients(messenger, addresses)

    @classmethod
    def get_subscribers(cls):
        """Returns a list of Recipient objects subscribed for this message type.

        :return:
        """
        subscribers_raw = Subscription.get_for_message_cls(cls.alias)
        subscribers = []

        for subscriber in subscribers_raw:
            messenger_cls = subscriber.messenger_cls
            address = subscriber.address
            recipient = subscriber.recipient
            if address is None:
                try:
                    address = get_registered_messenger_object(messenger_cls).get_address(recipient)
                except UnknownMessengerError:
                    pass

            if address and isinstance(address, string_types):
                subscribers.append(Recipient(messenger_cls, recipient, address))

        return subscribers

    @classmethod
    def get_dispatch_hash(cls, dispatch_id, message_id):
        """Returns a hash string for validation purposes.

        :param int dispatch_id:
        :param int message_id:
        :return:
        """
        return salted_hmac('%s' % dispatch_id, '%s|%s' % (message_id, dispatch_id)).hexdigest()

    @classmethod
    def get_mark_read_directive(cls, message_model, dispatch_model):
        """Returns mark read directive (command, URL, etc.) string.

        :param Message message_model:
        :param Dispatch dispatch_model:
        :return:
        """
        return cls._get_url('sitemessage_mark_read', message_model, dispatch_model)

    @classmethod
    def get_unsubscribe_directive(cls, message_model, dispatch_model):
        """Returns an unsubscribe directive (command, URL, etc.) string.

        :param Message message_model:
        :param Dispatch dispatch_model:
        :return:
        """
        return cls._get_url('sitemessage_unsubscribe', message_model, dispatch_model)

    @classmethod
    def _get_url(cls, name, message_model, dispatch_model):
        """Returns a common pattern sitemessage URL.

        :param str name: URL name
        :param Message message_model:
        :param Dispatch|None dispatch_model:
        :return:
        """
        global APP_URLS_ATTACHED

        url = ''

        if dispatch_model is None:
            return url

        if APP_URLS_ATTACHED != False:  # sic!

            hashed = cls.get_dispatch_hash(dispatch_model.id, message_model.id)

            try:
                url = reverse(name, args=[message_model.id, dispatch_model.id, hashed])
                url = '%s%s' % (get_site_url(), url)
            except NoReverseMatch:
                if APP_URLS_ATTACHED is None:
                    APP_URLS_ATTACHED = False

        return url

    @classmethod
    def handle_unsubscribe_request(cls, request, message, dispatch, hash_is_valid, redirect_to):
        """Handles user subscription cancelling request.

        :param Request request: Request instance
        :param Message message: Message model instance
        :param Dispatch dispatch: Dispatch model instance
        :param bool hash_is_valid: Flag indicating that user supplied request signature is correct
        :param str redirect_to: Redirection URL
        :rtype: list
        """

        if hash_is_valid:
            Subscription.cancel(
                dispatch.recipient_id or dispatch.address, cls.alias, dispatch.messenger
            )
            signal = sig_unsubscribe_success
        else:
            signal = sig_unsubscribe_failed

        signal.send(cls, request=request, message=message, dispatch=dispatch)
        return redirect(redirect_to)

    @classmethod
    def handle_mark_read_request(cls, request, message, dispatch, hash_is_valid, redirect_to):
        """Handles a request to mark a message as read.

        :param Request request: Request instance
        :param Message message: Message model instance
        :param Dispatch dispatch: Dispatch model instance
        :param bool hash_is_valid: Flag indicating that user supplied request signature is correct
        :param str redirect_to: Redirection URL
        :rtype: list
        """

        if hash_is_valid:
            dispatch.mark_read()
            dispatch.save()
            signal = sig_mark_read_success
        else:
            signal = sig_mark_read_failed

        signal.send(cls, request=request, message=message, dispatch=dispatch)
        return redirect(redirect_to)

    @classmethod
    def get_template(cls, message, messenger):
        """Get a template path to compile a message.

        1. `tpl` field of message context;
        2. `template` field of message class;
        3. deduced from message, messenger data and `template_ext` message type field
           (e.g. `sitemessage/messages/plain__smtp.txt` for `plain` message type).

        :param Message message: Message model
        :param MessengerBase messenger: a MessengerBase heir
        :return: str
        :rtype: str
        """
        template = message.context.get('tpl', None)

        if template:  # Template name is taken from message context.
            return template

        if cls.template is None:
            cls.template = 'sitemessage/messages/%s__%s.%s' % (
                cls.get_alias(), messenger.get_alias(), cls.template_ext
            )
        return cls.template

    @classmethod
    def compile(cls, message, messenger, dispatch=None):
        """Compiles and returns a message text.

        Considers `use_tpl` field from message context to decide whether
        template compilation is used.

        Otherwise a SIMPLE_TEXT_ID field from message context is used as message contents.

        :param Message message: model instance
        :param MessengerBase messenger: MessengerBase heir instance
        :param Dispatch dispatch: model instance to consider context from
        :return: str
        :rtype: str
        """
        if message.context.get('use_tpl', False):
            context = message.context
            context.update({
                'SITE_URL': get_site_url(),
                'directive_unsubscribe': cls.get_unsubscribe_directive(message, dispatch),
                'directive_mark_read': cls.get_mark_read_directive(message, dispatch),
                'message_model': message,
                'dispatch_model': dispatch
            })
            context = cls.get_template_context(context)
            return render_to_string(cls.get_template(message, messenger), context)
        return message.context[cls.SIMPLE_TEXT_ID]

    @classmethod
    def get_template_context(cls, context):
        """Returns context dict for template compilation.

        This method might be reimplemented by a heir to add some data into
        context before template compilation.

        :param dict context: Initial context
        :return:
        """
        return context

    @classmethod
    def update_context(cls, base_context, str_or_dict, template_path=None):
        """Helper method to structure initial message context data.

        NOTE: updates `base_context` inplace.

        :param dict base_context: context dict to update
        :param dict, str str_or_dict: text representing a message, or a dict to be placed into message context.
        :param str template_path: template path to be used for message rendering
        """
        if isinstance(str_or_dict, dict):
            base_context.update(str_or_dict)
            base_context['use_tpl'] = True
        else:
            base_context[cls.SIMPLE_TEXT_ID] = str_or_dict

        if cls.SIMPLE_TEXT_ID in str_or_dict:
            base_context['use_tpl'] = False

        base_context['tpl'] = template_path

    @classmethod
    def prepare_dispatches(cls, message, recipients=None):
        """Creates Dispatch models for a given message and return them.

        :param Message message: Message model instance
        :param list|None recipients: A list or Recipient objects
        :return: list of created Dispatch models
        :rtype: list
        """
        return Dispatch.create(message, recipients or cls.get_subscribers())



from django.conf import settings
from django.utils.translation import ugettext as _
from django.utils.html import strip_tags

from .base import MessengerBase
from ..exceptions import MessengerWarmupException


class SMTPMessenger(MessengerBase):
    """Implements SMTP message delivery using Python builtin smtplib module."""

    alias = 'smtp'
    smtp = None
    title = _('E-mail')
    _session_started = False

    def __init__(self, from_email=None, login=None, password=None, host=None, port=None, use_tls=None, use_ssl=None, debug=False):
        """Configures messenger.

        :param str from_email: e-mail address to send messages from
        :param str login: login to log into SMTP server
        :param str password: password to log into SMTP server
        :param str host: string - SMTP server host
        :param int port:  SMTP server port
        :param bool use_tls: whether to use TLS
        :param bool use_ssl: whether to use SSL
        :param bool debug: whether to switch smtplib into debug mode.
        """
        import smtplib

        from email.mime.text import MIMEText
        from email.mime.multipart import MIMEMultipart

        self.debug = debug
        self.lib = smtplib
        self.mime_text = MIMEText
        self.mime_multipart = MIMEMultipart

        self.from_email = from_email or getattr(settings, 'SERVER_EMAIL')
        self.login = login or getattr(settings, 'EMAIL_HOST_USER')
        self.password = password or getattr(settings, 'EMAIL_HOST_PASSWORD')
        self.host = host or getattr(settings, 'EMAIL_HOST')
        self.port = port or getattr(settings, 'EMAIL_PORT')
        self.use_tls = use_tls or getattr(settings, 'EMAIL_USE_TLS')
        self.use_ssl = use_ssl or getattr(settings, 'EMAIL_USE_SSL', False)  # False as default to support Django < 1.7

    @classmethod
    def get_address(cls, recipient):
        return getattr(recipient, 'email', None) or recipient

    def _test_message(self, to, text):
        return self._send_message(self._build_message(to, text, mtype='html'))

    def before_send(self):
        try:
            self.smtp = self.lib.SMTP(self.host, self.port)
            self.smtp.set_debuglevel(self.debug)

            if self.use_tls:
                self.smtp.ehlo()
                if self.smtp.has_extn('STARTTLS'):
                    self.smtp.starttls()
                    self.smtp.ehlo()  # This time over TLS.

            if self.login:
                self.smtp.login(self.login, self.password)

            self._session_started = True
        except self.lib.SMTPException as e:
            raise MessengerWarmupException('SMTP Error: %s' % e)

    def after_send(self):
        self.smtp.quit()

    def _build_message(self, to, text, subject=None, mtype=None, unsubscribe_url=None):
        """Constructs a MIME message from message and dispatch models."""
        # TODO Maybe file attachments handling through `files` message_model context var.

        if subject is None:
            subject = u'%s' % _('No Subject')

        if mtype == 'html':
            msg = self.mime_multipart()
            text_part = self.mime_multipart('alternative')
            text_part.attach(self.mime_text(strip_tags(text), _charset='utf-8'))
            text_part.attach(self.mime_text(text, 'html', _charset='utf-8'))
            msg.attach(text_part)
        else:
            msg = self.mime_text(text, _charset='utf-8')

        msg['From'] = self.from_email
        msg['To'] = to
        msg['Subject'] = subject

        if unsubscribe_url:
            msg['List-Unsubscribe'] = '<%s>' % unsubscribe_url

        return msg

    def _send_message(self, msg):
        return self.smtp.sendmail(msg['From'], msg['To'], msg.as_string())

    def send(self, message_cls, message_model, dispatch_models):
        if self._session_started:
            for dispatch_model in dispatch_models:

                msg = self._build_message(
                    dispatch_model.address,
                    dispatch_model.message_cache,
                    message_model.context.get('subject'),
                    message_model.context.get('type'),
                    message_cls.get_unsubscribe_directive(message_model, dispatch_model)
                )

                try:
                    refused = self._send_message(msg)
                    if refused:
                        self.mark_failed(dispatch_model, '`%s` address is rejected by server' % msg['To'])
                        continue
                    self.mark_sent(dispatch_model)
                except Exception as e:
                    self.mark_error(dispatch_model, e, message_cls)



import json

from django.core import exceptions
from django.db import models
from django.utils import timezone
from django.utils.six import with_metaclass, string_types
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import python_2_unicode_compatible
from django.conf import settings


USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


# This allows South to handle our custom 'CharFieldNullable' field.
if 'south' in settings.INSTALLED_APPS:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], ['^sitemessage\.models\.ContextField'])


class ContextField(with_metaclass(models.SubfieldBase, models.TextField)):

    def to_python(self, value):
        if not value:
            return {}

        if isinstance(value, dict):
            return value

        try:
            return json.loads(value)
        except ValueError:
            raise exceptions.ValidationError(_('Value `%r` is not a valid context.') % value,
                                             code='invalid_context', params={'value': value},)

    def get_prep_value(self, value):
        return json.dumps(value)


@python_2_unicode_compatible
class Message(models.Model):

    time_created = models.DateTimeField(_('Time created'), auto_now_add=True, editable=False)
    sender = models.ForeignKey(USER_MODEL, verbose_name=_('Sender'), null=True, blank=True)
    cls = models.CharField(_('Message class'), max_length=250, db_index=True,
                           help_text=_('Message logic class identifier.'))
    context = ContextField(_('Message context'))
    priority = models.PositiveIntegerField(_('Priority'), default=0, db_index=True,
                                           help_text=_('Number describing message sending priority. Messages with '
                                                       'different priorities can be sent with different periodicity.'))
    dispatches_ready = models.BooleanField(_('Dispatches ready'), db_index=True, default=False,
                                           help_text=_('Indicates whether dispatches for this message are already '
                                                       'formed and ready to delivery.'))

    class Meta:
        verbose_name = _('Message')
        verbose_name_plural = _('Messages')

    def __str__(self):
        return self.cls

    def get_type(self):
        """Returns message type (class) associated with the message.

        :raises UnknownMessageTypeError:
        """
        from .toolbox import get_registered_message_type
        return get_registered_message_type(self.cls)

    @classmethod
    def get_without_dispatches(cls):
        """Returns messages with no dispatches created.

        :return:
        """
        return cls.objects.filter(dispatches_ready=False).all()

    @classmethod
    def create(cls, message_class, context, recipients=None, sender=None, priority=None):
        """Creates a message (and dispatches).

        Returns a tuple: (message_model, list_of_dispatches)

        :param str message_class: alias of MessageBase heir
        :param dict context: context for a message
        :param list|None recipients: recipient (or a list) or None.
            If `None` Dispatches should be created before send using `prepare_dispatches()`.
        :param User|None sender: Django User model heir instance
        :param int|None priority: number describing message priority
        :return: a tuple with message model and a list of dispatch models.
        :rtype: tuple
        """
        dispatches_ready = False
        if recipients is not None:
            dispatches_ready = True

        msg_kwargs = {
            'cls': message_class,
            'context': context,
            'sender': sender,
            'dispatches_ready': dispatches_ready
        }

        if priority is not None:
            msg_kwargs['priority'] = priority

        message_model = cls(**msg_kwargs)
        message_model.save()
        dispatch_models = Dispatch.create(message_model, recipients)
        return message_model, dispatch_models


@python_2_unicode_compatible
class Dispatch(models.Model):

    DISPATCH_STATUS_PENDING = 1
    DISPATCH_STATUS_SENT = 2
    DISPATCH_STATUS_ERROR = 3
    DISPATCH_STATUS_FAILED = 4

    DISPATCH_STATUSES = (
        (DISPATCH_STATUS_PENDING, _('Pending')),
        (DISPATCH_STATUS_SENT, _('Sent')),
        (DISPATCH_STATUS_ERROR, _('Error')),
        (DISPATCH_STATUS_FAILED, _('Failed')),
    )

    READ_STATUS_UNDREAD = 0
    READ_STATUS_READ = 1

    READ_STATUSES = (
        (READ_STATUS_UNDREAD, _('Unread')),
        (READ_STATUS_READ, _('Read')),
    )

    error_log = None

    time_created = models.DateTimeField(_('Time created'), auto_now_add=True, editable=False)
    time_dispatched = models.DateTimeField(_('Time dispatched'), editable=False, null=True, blank=True,
                                           help_text=_('Time of the last delivery attempt.'))

    message = models.ForeignKey(Message, verbose_name=_('Message'))
    messenger = models.CharField(_('Messenger'), max_length=250, db_index=True,
                                 help_text=_('Messenger class identifier.'))

    recipient = models.ForeignKey(USER_MODEL, verbose_name=_('Recipient'), null=True, blank=True)
    address = models.CharField(_('Address'), max_length=250, help_text=_('Recipient address.'))
    retry_count = models.PositiveIntegerField(_('Retry count'), default=0,
                                              help_text=_('A number of delivery retries has already been made.'))

    message_cache = models.TextField(_('Message cache'), null=True, editable=False)
    dispatch_status = models.PositiveIntegerField(_('Dispatch status'), choices=DISPATCH_STATUSES,
                                                  default=DISPATCH_STATUS_PENDING)

    read_status = models.PositiveIntegerField(_('Read status'), choices=READ_STATUSES, default=READ_STATUS_UNDREAD)

    class Meta:
        verbose_name = _('Dispatch')
        verbose_name_plural = _('Dispatches')

    def __str__(self):
        return '%s [%s]' % (self.address, self.messenger)

    def is_read(self):
        """Returns message read flag.

        :rtype: bool
        :return:
        """
        return self.read_status == self.READ_STATUS_READ

    def mark_read(self):
        """Marks message as read (doesn't save it).

        :return:
        """
        self.read_status = self.READ_STATUS_READ

    @classmethod
    def log_dispatches_errors(cls, dispatches):
        """Batch logs dispatches delivery errors into DB.

        :param list dispatches:
        :return:
        """
        error_entries = []
        for dispatch in dispatches:
            # Saving message body cache for further usage.
            dispatch.save()
            error_entries.append(DispatchError(dispatch=dispatch, error_log=dispatch.error_log))
        DispatchError.objects.bulk_create(error_entries)

    @classmethod
    def set_dispatches_statuses(cls, **statuses):
        """Batch set dispatches delivery statuses using a [kwargs] dictionary
        of dispatch lists indexed by statuses.

        :param statuses:
        :return:
        """
        kwarg_status_map = {
            'sent': cls.DISPATCH_STATUS_SENT,
            'error': cls.DISPATCH_STATUS_ERROR,
            'failed': cls.DISPATCH_STATUS_FAILED,
            'pending': cls.DISPATCH_STATUS_PENDING,
        }
        for status_name, real_status in kwarg_status_map.items():
            if statuses.get(status_name, False):
                update_kwargs = {
                    'time_dispatched': timezone.now(),
                    'dispatch_status': real_status,
                    'retry_count': models.F('retry_count') + 1
                }
                cls.objects.filter(id__in=[d.id for d in statuses[status_name]]).update(**update_kwargs)

    @staticmethod
    def group_by_messengers(dispatches):
        """Groups dispatches by messages.

        :param list dispatches:
        :return:
        """
        by_messengers = {}
        for dispatch in dispatches:
            if dispatch.messenger not in by_messengers:
                by_messengers[dispatch.messenger] = {}

            if dispatch.message.id not in by_messengers[dispatch.messenger]:
                by_messengers[dispatch.messenger][dispatch.message.id] = (dispatch.message, [])

            by_messengers[dispatch.messenger][dispatch.message.id][1].append(dispatch)
        return by_messengers

    @classmethod
    def get_unsent(cls, priority=None):
        """Returns dispatches unsent (scheduled or with errors).

        :param int priority: Message priority filter
        :return:
        """
        filter_kwargs = {
            'dispatch_status__in': (cls.DISPATCH_STATUS_PENDING, cls.DISPATCH_STATUS_ERROR)
        }
        if priority is not None:
            filter_kwargs['message__priority'] = priority
        return cls.objects.select_related('message').filter(**filter_kwargs).order_by('-message__time_created').all()

    @classmethod
    def get_unread(cls):
        """Returns unread dispatches.

        :return:
        """
        return cls.objects.filter(read_status=cls.READ_STATUS_UNDREAD).select_related('message').all()

    @classmethod
    def create(cls, message_model, recipients):
        """Creates dispatches for given recipients.

        NB: dispatch models are bulk created and do not have IDs.

        :param Message message_model:
        :param recipients:
        :return:
        """
        objects = []
        if recipients:
            if not isinstance(recipients, (list, set)):
                recipients = (recipients,)

            for r in recipients:
                objects.append(cls(message=message_model, messenger=r.messenger, recipient=r.user, address=r.address))

            if objects:
                cls.objects.bulk_create(objects)

            if not message_model.dispatches_ready:
                message_model.dispatches_ready = True
                message_model.save()

        return objects


@python_2_unicode_compatible
class DispatchError(models.Model):

    time_created = models.DateTimeField(_('Time created'), auto_now_add=True, editable=False)
    dispatch = models.ForeignKey(Dispatch, verbose_name=_('Dispatch'))
    error_log = models.TextField(_('Text'))

    class Meta:
        verbose_name = _('Dispatch error')
        verbose_name_plural = _('Dispatch errors')

    def __str__(self):
        return 'Dispatch ID %s error entry' % self.dispatch_id


@python_2_unicode_compatible
class Subscription(models.Model):

    time_created = models.DateTimeField(_('Time created'), auto_now_add=True, editable=False)
    message_cls = models.CharField(_('Message class'), max_length=250, db_index=True,
                                   help_text=_('Message logic class identifier.'))
    messenger_cls = models.CharField(_('Messenger'), max_length=250, db_index=True,
                                     help_text=_('Messenger class identifier.'))
    recipient = models.ForeignKey(USER_MODEL, verbose_name=_('Recipient'), null=True, blank=True)
    address = models.CharField(_('Address'), max_length=250, null=True, help_text=_('Recipient address.'))

    class Meta:
        verbose_name = _('Subscription')
        verbose_name_plural = _('Subscriptions')

    def __str__(self):
        recipient = self.recipient_id or self.address
        return '%s [%s - %s]' % (recipient, self.message_cls, self.messenger_cls)

    @classmethod
    def get_for_user(cls, user):
        """Returns subscriptions for a given user.

        :param User user:
        :return:
        """
        if user.id is None:
            return []

        return cls.objects.filter(recipient=user)

    @classmethod
    def replace_for_user(cls, user, prefs):
        """Set subscription preferences for a given user.

        :param User user:
        :param list prefs: List of tuples (message_cls, messenger_cls)
        :return:
        """
        uid = user.id
        if uid is None:
            return False

        # Remove previous prefs.
        cls.objects.filter(recipient_id=uid).delete()

        new_prefs = []

        for pref in prefs:
            new_prefs.append(
                cls(**cls._get_base_kwargs(uid, pref[0], pref[1]))
            )

        if new_prefs:
            cls.objects.bulk_create(new_prefs)

        return True

    @classmethod
    def get_for_message_cls(cls, message_cls):
        """Returns subscriptions for a given message class alias.

        :param str message_cls:
        :return:
        """
        return cls.objects.select_related('recipient').filter(message_cls=message_cls)

    @classmethod
    def _get_base_kwargs(cls, recipient, message_cls, messenger_cls):

        if not isinstance(message_cls, string_types):
            message_cls = message_cls.alias

        if not isinstance(messenger_cls, string_types):
            messenger_cls = messenger_cls.alias

        base_kwargs = {
            'message_cls': message_cls,
            'messenger_cls': messenger_cls,
        }
        if isinstance(recipient, string_types):
            base_kwargs['address'] = recipient
        else:
            if not isinstance(recipient, int):
                recipient = recipient.id
            base_kwargs['recipient_id'] = recipient

        return base_kwargs

    @classmethod
    def create(cls, uid_or_address, message_cls, messenger_cls):
        """Creates a subscription for a recipient.

        :param int|str uid_or_address: User ID or address string.
        :param str|MessageBase message_cls: Message type alias or class
        :param str|MessengerBase messenger_cls: Messenger type alias or class
        :return:
        """
        obj = cls(**cls._get_base_kwargs(uid_or_address, message_cls, messenger_cls))
        obj.save()
        return obj

    @classmethod
    def cancel(cls, uid_or_address, message_cls, messenger_cls):
        """Cancels a subscription for a recipient.

        :param int|str uid_or_address: User ID or address string.
        :param str|MessageBase message_cls: Message type alias or class
        :param str|MessengerBase messenger_cls: Messenger type alias or class
        :return:
        """
        cls.objects.filter(
            **cls._get_base_kwargs(uid_or_address, message_cls, messenger_cls)
        ).delete()



import os
from setuptools import setup
from sitemessage import VERSION


PATH_BASE = os.path.dirname(__file__)

f = open(os.path.join(PATH_BASE, 'README.rst'))
README = f.read()
f.close()

setup(
    name='django-sitemessage',
    version='.'.join(map(str, VERSION)),
    url='https://github.com/idlesign/django-sitemessage',

    description='Reusable application for Django introducing a message delivery framework',
    long_description=README,
    license='BSD 3-Clause License',

    author='Igor `idle sign` Starikov',
    author_email='idlesign@yandex.ru',

    packages=['sitemessage'],
    include_package_data=True,
    zip_safe=False,

    install_requires=['django-etc'],

    classifiers=[
        # As in https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'License :: OSI Approved :: BSD License'
    ],
)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('sitemessage', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('time_created', models.DateTimeField(auto_now_add=True, verbose_name='Time created')),
                ('message_cls', models.CharField(help_text='Message logic class identifier.', max_length=250, verbose_name='Message class', db_index=True)),
                ('messenger_cls', models.CharField(help_text='Messenger class identifier.', max_length=250, verbose_name='Messenger', db_index=True)),
                ('address', models.CharField(help_text='Recipient address.', max_length=250, null=True, verbose_name='Address')),
                ('recipient', models.ForeignKey(verbose_name='Recipient', blank=True, to=settings.AUTH_USER_MODEL, null=True)),
            ],
            options={
                'verbose_name': 'Subscription',
                'verbose_name_plural': 'Subscriptions',
            },
            bases=(models.Model,),
        ),
    ]




class SiteMessageError(Exception):
    """Base class for sitemessage errors."""


class SiteMessageConfigurationError(SiteMessageError):
    """This error is raised on configuration errors."""


class UnknownMessageTypeError(SiteMessageError):
    """This error is raised when there's a try to access an unknown message type."""


class MessengerException(SiteMessageError):
    """Base messenger exception."""


class UnknownMessengerError(MessengerException):
    """This error is raised when there's a try to access an unknown messenger."""


class MessengerWarmupException(MessengerException):
    """This exception represents a delivery error due to a messenger warm up process failure."""



from collections import namedtuple, defaultdict, OrderedDict
from threading import local

from django.utils import six

try:
    from django.utils.module_loading import import_module
except ImportError:
    # Django <=1.9.0
    from django.utils.importlib import import_module

from django.utils.module_loading import module_has_submodule
from etc.toolbox import get_site_url as get_site_url_

from .settings import APP_MODULE_NAME, SITE_URL
from .exceptions import UnknownMessageTypeError, UnknownMessengerError


_MESSENGERS_REGISTRY = OrderedDict()
_MESSAGES_REGISTRY = OrderedDict()

_MESSAGES_FOR_APPS = defaultdict(dict)

_THREAD_LOCAL = local()
_THREAD_SITE_URL = 'sitemessage_site_url'


def get_site_url():
    """Returns a URL for current site.

    :return:
    """

    site_url = getattr(_THREAD_LOCAL, _THREAD_SITE_URL, None)
    if site_url is None:
        site_url = SITE_URL or get_site_url_()
        setattr(_THREAD_LOCAL, _THREAD_SITE_URL, site_url)

    return site_url


def get_message_type_for_app(app_name, default_message_type_alias):
    """Returns a registered message type object for a given application.

    Supposed to be used by reusable applications authors,
    to get message type objects which may be overridden by project authors
    using `override_message_type_for_app`.

    :param str app_name:
    :param str default_message_type_alias:
    :return: a message type object overridden is so, or the default
    :rtype: MessageBase
    """
    message_type = default_message_type_alias
    try:
        message_type = _MESSAGES_FOR_APPS[app_name][message_type]
    except KeyError:
        pass
    return get_registered_message_type(message_type)


def override_message_type_for_app(app_name, app_message_type_alias, new_message_type_alias):
    """Overrides a given message type used by a certain application with another one.

    Intended for projects authors, who need to customize messaging behaviour
    of a certain thirdparty app (supporting this feature).
    To be used in conjunction with `get_message_type_for_app`.

    :param str app_name:
    :param str app_message_type_alias:
    :param str new_message_type_alias:
    :return:
    """
    global _MESSAGES_FOR_APPS

    _MESSAGES_FOR_APPS[app_name][app_message_type_alias] = new_message_type_alias


def register_messenger_objects(*messengers):
    """Registers (configures) messengers.

    :param list messengers: MessengerBase heirs instances.
    """
    global _MESSENGERS_REGISTRY

    for messenger in messengers:
        _MESSENGERS_REGISTRY[messenger.get_alias()] = messenger


def get_registered_messenger_objects():
    """Returns registered (configured) messengers dict
    indexed by messenger aliases.

    :return: dict
    :rtype: dict
    """
    return _MESSENGERS_REGISTRY


def get_registered_messenger_object(messenger):
    """Returns registered (configured) messenger by alias,

    :param str messenger: messenger alias
    :return: MessengerBase heirs instances.
    :rtype: MessengerBase
    """
    try:
        return _MESSENGERS_REGISTRY[messenger]
    except KeyError:
        raise UnknownMessengerError('`%s` messenger is not registered' % messenger)


def register_message_types(*message_types):
    """Registers message types (classes).

    :param list message_types: MessageBase heir classes.
    """
    global _MESSAGES_REGISTRY

    for message in message_types:
        _MESSAGES_REGISTRY[message.get_alias()] = message


def get_registered_message_types():
    """Returns registered message types dict indexed by their aliases.

    :return: dict
    :rtype: dict
    """
    return _MESSAGES_REGISTRY


def get_registered_message_type(message_type):
    """Returns registered message type (class) by alias,

    :param str message_type: message type alias
    :return: MessageBase heirs instances.
    :rtype: MessageBase
    """
    try:
        return _MESSAGES_REGISTRY[message_type]
    except KeyError:
        raise UnknownMessageTypeError('`%s` message class is not registered' % message_type)


def import_app_sitemessage_module(app):
    """Returns a submodule of a given app

    :param str app: application name
    :return: submodule or None
    :rtype: module or None
    """
    module_name = APP_MODULE_NAME
    module = import_module(app)
    try:
        sub_module = import_module('%s.%s' % (app, module_name))
        return sub_module
    except:
        if module_has_submodule(module, module_name):
            raise
        return None


def import_project_sitemessage_modules():
    """Imports sitemessages modules from registered apps."""
    from django.conf import settings as django_settings
    submodules = []
    for app in django_settings.INSTALLED_APPS:
        module = import_app_sitemessage_module(app)
        if module is not None:
            submodules.append(module)
    return submodules


def is_iterable(v):
    """Tells whether the thing is an iterable.
    NB: strings do not count even on Py3.

    """
    return hasattr(v, '__iter__') and not isinstance(v, six.string_types)


# Class used to represent message recipients.
Recipient = namedtuple('Recipient', ('messenger', 'user', 'address'))


def recipients(messenger, addresses):
    """Structures recipients data.

    :param str, MessageBase messenger: MessengerBase heir
    :param list addresses: recipients addresses or Django User model heir instances (NOTE: if supported by a messenger)
    :return: list of Recipient
    :rtype: list
    """
    if isinstance(messenger, six.string_types):
        messenger = get_registered_messenger_object(messenger)
    return messenger._structure_recipients_data(addresses)



from django.utils.translation import ugettext as _

from .base import MessengerBase
from ..exceptions import MessengerWarmupException


class TwitterMessenger(MessengerBase):
    """Implements to Twitter message delivery using `twitter` module.

    https://github.com/sixohsix/twitter

    """

    alias = 'twitter'
    title = _('Tweet')
    _session_started = False

    def __init__(self, api_key, api_secret, access_token, access_token_secret):
        """Configures messenger.

        Register Twitter application here - https://apps.twitter.com/

        :param api_key: API key for Twitter client
        :param api_secret: API secret for Twitter client
        :param access_token: Access token for an account to tweet from
        :param access_token_secret: Access token secret for an account to tweet from
        """
        import twitter

        self.lib = twitter
        self.api_key = api_key
        self.api_secret = api_secret
        self.access_token = access_token
        self.access_token_secret = access_token_secret

    @classmethod
    def get_address(cls, recipient):
        return getattr(recipient, 'twitter', None) or recipient

    def _test_message(self, to, text):
        return self._send_message(self._build_message(to, text))

    def before_send(self):
        try:
            self.api = self.lib.Twitter(auth=self.lib.OAuth(self.access_token, self.access_token_secret, self.api_key, self.api_secret))
            self._session_started = True
        except self.lib.api.TwitterError as e:
            raise MessengerWarmupException('Twitter Error: %s' % e)
    
    @classmethod
    def _build_message(cls, to, text):
        if to:
            if not to.startswith('@'):
                to = '@%s' % to
            to = '%s ' % to
        else:
            to = ''
        msg = '%s%s' % (to, text)
        return msg

    def _send_message(self, msg):
        return self.api.statuses.update(status=msg)

    def send(self, message_cls, message_model, dispatch_models):
        if self._session_started:
            for dispatch_model in dispatch_models:
                msg = self._build_message(dispatch_model.address, dispatch_model.message_cache)
                try:
                    self._send_message(msg)
                    self.mark_sent(dispatch_model)
                except Exception as e:
                    self.mark_error(dispatch_model, e, message_cls)
